import inspect
import os
import re
import sys
import sysconfig
from pathlib import Path
from subprocess import PIPE
from subprocess import run

import pytest
from pytest_mock import MockerFixture

from lektor.environment import Environment
from lektor.packages import load_packages
from lektor.packages import Requirements
from lektor.packages import update_cache
from lektor.packages import VirtualEnv
from lektor.project import Project


@pytest.fixture(scope="module")
def nested_venv(tmp_path_factory: pytest.TempPathFactory) -> VirtualEnv:
    """Create a lighweight nested virtual environment.

    The created venv does not have anything installed in it — not even pip. It does,
    however, have our ``site-packages`` directory added to its ``sys.path``, so it
    should have access to a working ``pip`` that way.

    This lightweight venv is relative quick to create.  Creating a full independent
    venv involves running ``python -m ensurepip`` and potential network requests
    to PyPI.
    """
    tmp_path = tmp_path_factory.mktemp("nested_venv")
    venv = VirtualEnv(tmp_path)
    # venv creation is very quick without installing/upgrading pip
    venv.create(with_pip=False, upgrade_deps=False)

    # add our site-packages to the venv's sys.path
    venv.addsitedir(sysconfig.get_path("purelib"))
    return venv


def test_VirtualEnv_creates_site_packages(tmp_path: Path) -> None:
    venv = VirtualEnv(tmp_path)
    # venv creation is very quick without installing/upgrading pip
    venv.create(with_pip=False, upgrade_deps=False)
    assert Path(venv.site_packages).is_dir()


def test_VirtualEnv_addsitedir(nested_venv: VirtualEnv) -> None:
    # check that we can run pytest (presumably from our site-packages)
    proc = run((nested_venv.executable, "-m", "pytest", "--version"), check=False)
    assert proc.returncode == 0


@pytest.mark.skipif(os.name == "nt", reason="Windows")
def test_VirtualEnv_uses_symlinks(nested_venv: VirtualEnv) -> None:
    executable = Path(nested_venv.executable)
    assert executable.resolve(strict=True).parent != executable.parent


@pytest.mark.requiresinternet
@pytest.mark.slowtest
def test_VirtualEnv_run_pip_install(tmp_path: Path) -> None:
    # XXX: slow test
    venv = VirtualEnv(tmp_path)
    venv.create()

    # install a dummy plugin
    plugin_path = Path(__file__).parent / "pep660-dummy-plugin"
    dummy_plugin_path = os.fspath(plugin_path.resolve())
    venv.run_pip_install(f"--editable={dummy_plugin_path}")

    # Make our lektor available to the installed plugin
    venv.addsitedir(sysconfig.get_path("purelib"))

    # Check that we can load the plugin entry point
    prog = inspect.cleandoc(
        """
        import sys
        from importlib import metadata

        for ep in metadata.entry_points(group="lektor.plugins", name="dummy"):
            print(ep.load().__name__)
    """
    )
    proc = run((venv.executable, "-c", prog), stdout=PIPE, text=True, check=True)
    assert proc.stdout.strip() == "DummyPlugin"


def test_VirtualEnv_run_pip_install_raises_runtime_error(
    nested_venv: VirtualEnv, capfd: pytest.CaptureFixture[str]
) -> None:
    with pytest.raises(RuntimeError) as excinfo:
        nested_venv.run_pip_install("--unknown-option")
    assert excinfo.match("Failed to install")
    assert "no such option" in capfd.readouterr().err


def test_VirtualEnv_site_packages(tmp_path: Path) -> None:
    site_packages = VirtualEnv(tmp_path).site_packages
    relpath = os.fspath(Path(site_packages).relative_to(tmp_path))
    assert re.match(r"(?i)lib(?=[/\\]).*[/\\]site-packages\Z", relpath)


def test_VirtualEnv_executable(tmp_path: Path) -> None:
    executable = Path(VirtualEnv(tmp_path).executable)
    scripts_dir = os.fspath(executable.parent.relative_to(tmp_path))
    assert scripts_dir in {"bin", "Scripts"}
    assert executable.name == Path(sys.executable).name


def test_Requirements_add_requirement() -> None:
    requirements = Requirements()
    requirements.add_requirement("foo", "1.2")
    requirements.add_requirement("bar")
    assert len(requirements) == 2
    assert set(requirements) == {"foo==1.2", "bar"}


def test_Requirements_add_local_requirement() -> None:
    requirements = Requirements()
    plugin_path = Path(__file__).parent / "pep660-dummy-plugin"
    requirements.add_local_requirement(plugin_path)
    assert set(requirements) == {f"--editable={os.fspath(plugin_path.resolve())}"}


def test_Requirements_add_local_requirements_from(tmp_path: Path) -> None:
    for fn in ["plugin1/pyproject.toml", "plugin2/setup.py", "notaplugin/README.md"]:
        path = tmp_path / fn
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()
    requirements = Requirements()
    requirements.add_local_requirements_from(tmp_path)
    assert len(requirements) == 2
    assert {req.rpartition(os.sep)[2] for req in requirements} == {"plugin1", "plugin2"}


def test_Requirements_add_local_requirements_from_missing_dir(tmp_path: Path) -> None:
    requirements = Requirements()
    requirements.add_local_requirements_from(tmp_path / "missing")
    assert len(requirements) == 0
    assert not requirements


def test_Requirements_hash() -> None:
    requirements = Requirements()
    assert requirements.hash() == "da39a3ee5e6b4b0d3255bfef95601890afd80709"
    requirements.add_requirement("foo", "42")
    assert requirements.hash() == "a44f078eab8bc1aa1ddfd111d63e24ff65131b4b"


def test_update_cache_installs_requirements(
    tmp_path: Path, mocker: MockerFixture
) -> None:
    venv_path = tmp_path / "cache"
    venv_path.mkdir()
    VirtualEnv = mocker.patch("lektor.packages.VirtualEnv")
    update_cache(venv_path, {"foo": "42"}, tmp_path / "packages")
    assert mocker.call().run_pip_install("foo==42") in VirtualEnv.mock_calls
    hash_file = venv_path / "lektor-requirements-hash.txt"
    assert hash_file.read_text().strip() == "a44f078eab8bc1aa1ddfd111d63e24ff65131b4b"


def test_update_cache_skips_install_if_up_to_date(
    tmp_path: Path, mocker: MockerFixture
) -> None:
    venv_path = tmp_path / "cache"
    venv_path.mkdir()
    venv_path.joinpath("lektor-requirements-hash.txt").write_text(
        "a44f078eab8bc1aa1ddfd111d63e24ff65131b4b\n"
    )
    VirtualEnv = mocker.patch("lektor.packages.VirtualEnv")
    update_cache(venv_path, {"foo": "42"}, tmp_path / "packages")
    assert VirtualEnv.mock_calls == []


def test_update_cache_removes_package_cache_if_no_requirements(tmp_path: Path) -> None:
    venv_path = tmp_path / "cache"
    venv_path.mkdir()

    update_cache(venv_path, {}, tmp_path / "missing")
    assert not venv_path.exists()


def test_load_packages_add_package_cache_to_sys_path(env: Environment) -> None:
    load_packages(env)
    venv_path = env.project.get_package_cache_path()
    site_packages = VirtualEnv(venv_path).site_packages
    assert site_packages in sys.path


PackageCacheType = Project.PackageCacheType


@pytest.mark.parametrize("cache_type", PackageCacheType)
def test_load_packages_reinstall_wipes_cache(
    env: Environment, cache_type: PackageCacheType
) -> None:
    project = env.project
    cache_path = project.get_package_cache_path(cache_type)
    cache_path.mkdir(parents=True, exist_ok=False)

    load_packages(env, reinstall=True)
    assert not cache_path.exists()



"""Unit tests for lektor.pluginsystem."""

from __future__ import annotations

import inspect
import sys
from importlib import metadata
from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from pathlib import Path
from unittest import mock

import pytest

from lektor.cli import cli
from lektor.context import Context
from lektor.packages import add_package_to_project
from lektor.pluginsystem import _check_dist_name
from lektor.pluginsystem import _find_plugins
from lektor.pluginsystem import get_plugin
from lektor.pluginsystem import Plugin
from lektor.pluginsystem import PluginController


class DummyPlugin(Plugin):
    name = "Dummy Plugin"
    description = "For testing"

    # DummyPlugin.calls is test-local (see fixture dummy_plugin_calls, below)
    calls = []

    def __getattr__(self, name):
        if not name.startswith("on_"):
            raise AttributeError(name)

        event = name[3:].replace("_", "-")

        def hook(extra_flags, **kwargs):
            call = {
                "event": event,
                "extra_flags": extra_flags,
                "kwargs": kwargs,
            }
            self.calls.append(call)
            return f"{event} return value"

        return hook

    def on_legacy_event(self):
        self.calls.append({"event": "legacy-event"})
        return "legacy-event return value"

    def on_one_type_error(self, **kwargs):
        """Raises TypeError only on the first call."""
        self.calls.append({"event": "one-type-error"})
        if len(self.calls) == 1:
            raise TypeError("test")
        return "one-type-error return value"


@pytest.fixture(autouse=True)
def dummy_plugin_calls(monkeypatch):
    """Reset list of DummyPlugin hook calls for each test."""
    monkeypatch.setattr(DummyPlugin, "calls", [])
    return DummyPlugin.calls


class DummyDistribution(metadata.Distribution):
    _files = {
        "top_level.txt": f"{__name__}\n",
        "entry_points.txt": inspect.cleandoc(
            f"""
            [lektor.plugins]
            dummy-plugin = {__name__}:DummyPlugin
            """
        ),
    }

    # Allow overriding inherited properties with class attributes
    metadata = None

    def __init__(self, metadata):
        self.metadata = metadata

    def read_text(self, filename):
        return self._files.get(filename)

    def locate_file(self, path):  # pylint: disable=no-self-use
        return None


class DummyPluginLoader(Loader):
    # pylint: disable=abstract-method
    # pylint: disable=no-self-use

    def create_module(self, spec):
        return None

    def exec_module(self, module):
        module.DummyPlugin = DummyPlugin


class DummyPluginFinder(metadata.DistributionFinder):
    def __init__(self, module: str, distribution: metadata.Distribution):
        self.module = module
        self.distribution = distribution

    def find_spec(self, fullname, path, target=None):
        if fullname == self.module and path is None:
            return ModuleSpec(fullname, DummyPluginLoader())
        return None

    def find_distributions(
        self, context: metadata.DistributionFinder.Context | None = None
    ):
        return [self.distribution]


@pytest.fixture
def dummy_plugin_distribution(save_sys_path):
    """Add a dummy plugin distribution to the current working_set."""
    dist = DummyDistribution(
        {
            "Name": "lektor-dummy-plugin",
            "Version": "1.23",
        }
    )
    finder = DummyPluginFinder(__name__, dist)
    # The save_sys_path fixture will restore meta_path at end of test
    sys.meta_path.insert(0, finder)
    return dist


@pytest.fixture
def dummy_plugin(env, dummy_plugin_distribution):
    """Instantiate and register a dummy plugin in env"""
    env.plugin_controller.instanciate_plugin(
        "dummy-plugin", DummyPlugin, dummy_plugin_distribution
    )
    return env.plugins["dummy-plugin"]


def test_get_plugin(env, dummy_plugin):
    assert get_plugin("dummy-plugin", env) == dummy_plugin


def test_get_plugin_from_context(env, dummy_plugin):
    with Context(pad=env.new_pad()):
        assert get_plugin("dummy-plugin") == dummy_plugin


def test_get_plugin_missing(env):
    with pytest.raises(LookupError, match=r"Plugin .* not found"):
        get_plugin("dummy-plugin", env)


def test_get_plugin_no_env_or_ctx(dummy_plugin):
    with pytest.raises(RuntimeError, match=r"Context is unavailable"):
        get_plugin("dummy-plugin")


class TestPlugin:
    # pylint: disable=no-self-use

    @pytest.fixture
    def scratch_project_data(self, scratch_project_data):
        """Add plugin config file to scratch project."""
        plugin_ini = scratch_project_data / "configs/dummy-plugin.ini"
        plugin_ini.parent.mkdir(exist_ok=True)
        plugin_ini.write_text("test_setting = test value\n", "utf-8")
        return scratch_project_data

    @pytest.fixture
    def scratch_plugin(self, scratch_env, dummy_plugin_distribution):
        """Instantiate and register a dummy plugin with scratch_env"""
        scratch_env.plugin_controller.instanciate_plugin(
            "dummy-plugin", DummyPlugin, dummy_plugin_distribution
        )
        return scratch_env.plugins["dummy-plugin"]

    def test_env(self, dummy_plugin, env):
        assert dummy_plugin.env is env

    def test_env_went_away(self):
        env = mock.Mock(name="env", spec=())
        plugin = DummyPlugin(env, "dummy-plugin")
        del env
        with pytest.raises(RuntimeError, match=r"Environment went away"):
            _ = plugin.env

    def test_version(self, dummy_plugin, dummy_plugin_distribution):
        assert dummy_plugin.version == dummy_plugin_distribution.version

    def test_version_missing(self, env):
        # Instantiate plugin without specifying distribution
        env.plugin_controller.instanciate_plugin("dummy-plugin", DummyPlugin)
        plugin = env.plugins["dummy-plugin"]
        assert plugin.version is None

    def test_path(self, dummy_plugin):
        assert dummy_plugin.path == str(Path(__file__).parent)

    @pytest.mark.requiresinternet
    @pytest.mark.slowtest
    @pytest.mark.usefixtures("save_sys_path")
    def test_path_installed_plugin_is_none(self, scratch_project):
        # XXX: this test is slow and fragile. (It won't run
        # without an internet connection.)
        add_package_to_project(scratch_project, "webpack-support")
        env = scratch_project.make_env()
        plugin = get_plugin("webpack-support", env)
        assert plugin.path is None

    def test_import_name(self, dummy_plugin):
        assert dummy_plugin.import_name == f"{__name__}:DummyPlugin"

    def test_get_lektor_config(self, dummy_plugin):
        cfg = dummy_plugin.get_lektor_config()
        assert cfg["PROJECT"]["name"] == "Demo Project"

    def test_get_lektor_config_from_context(self, dummy_plugin, env):
        with Context(pad=env.new_pad()):
            cfg = dummy_plugin.get_lektor_config()
        assert cfg["PROJECT"]["name"] == "Demo Project"

    def test_config_filename(self, dummy_plugin, env):
        configdir = Path(env.root_path) / "configs"
        assert dummy_plugin.config_filename == str(configdir / "dummy-plugin.ini")

    def test_get_config(self, scratch_plugin):
        cfg = scratch_plugin.get_config()
        assert cfg["test_setting"] == "test value"

    def test_get_config_records_dependency(self, scratch_plugin, scratch_env):
        with Context(pad=scratch_env.new_pad()) as ctx:
            scratch_plugin.get_config()
        assert scratch_plugin.config_filename in ctx.referenced_dependencies

    def test_get_config_returns_cached_value(self, scratch_plugin, scratch_env):
        with Context(pad=scratch_env.new_pad()):
            cfg = scratch_plugin.get_config()
            cfg["is_cached"] = "indeed"
            cfg = scratch_plugin.get_config()
        assert "is_cached" in cfg

    def test_get_config_fresh(self, scratch_plugin, scratch_env):
        with Context(pad=scratch_env.new_pad()):
            cfg = scratch_plugin.get_config()
            cfg["is_cached"] = "indeed"
            cfg = scratch_plugin.get_config(fresh=True)
        assert "is_cached" not in cfg

    def test_emit(self, dummy_plugin):
        rv = dummy_plugin.emit("subevent")
        assert rv == {"dummy-plugin": "dummy-plugin-subevent return value"}
        events = [call["event"] for call in dummy_plugin.calls]
        assert events == ["dummy-plugin-subevent"]

    def test_to_json(self, dummy_plugin, dummy_plugin_distribution):
        assert dummy_plugin.to_json() == {
            "id": "dummy-plugin",
            "name": DummyPlugin.name,
            "description": DummyPlugin.description,
            "version": dummy_plugin_distribution.version,
            "import_name": f"{__name__}:DummyPlugin",
            "path": str(Path(__file__).parent),
        }


def test_find_plugins(dummy_plugin_distribution):
    (ep,) = dummy_plugin_distribution.entry_points
    assert list(_find_plugins()) == [(dummy_plugin_distribution, ep)]


@pytest.mark.parametrize(
    "dist_name, plugin_id",
    [
        ("Lektor-FOO", "Foo"),
    ],
)
def test_check_dist_name(dist_name, plugin_id):
    _check_dist_name(dist_name, plugin_id)


@pytest.mark.parametrize(
    "dist_name, plugin_id",
    [
        ("NotLektor-FOO", "Foo"),
    ],
)
def test_check_dist_name_raises(dist_name, plugin_id):
    with pytest.raises(RuntimeError):
        _check_dist_name(dist_name, plugin_id)


class TestPluginController:
    # pylint: disable=no-self-use

    @pytest.fixture
    def extra_flags(self):
        return {"flag": "value"}

    @pytest.fixture
    def plugin_controller(self, env, extra_flags):
        return PluginController(env, extra_flags)

    def test_env(self, plugin_controller, env):
        assert plugin_controller.env is env

    def test_env_went_away(self):
        env = mock.Mock(name="env", spec=())
        plugin_controller = PluginController(env)
        del env
        with pytest.raises(RuntimeError, match=r"Environment went away"):
            _ = plugin_controller.env

    def test_instantiate_plugin(self, plugin_controller, env):
        plugin_controller.instanciate_plugin("plugin-id", DummyPlugin)
        assert isinstance(env.plugins["plugin-id"], DummyPlugin)
        assert env.plugin_ids_by_class[DummyPlugin] == "plugin-id"

    def test_instantiate_plugin_error(self, plugin_controller, dummy_plugin):
        with pytest.raises(RuntimeError, match=r"already registered"):
            plugin_controller.instanciate_plugin(dummy_plugin.id, DummyPlugin)

    def test_iter_plugins(self, plugin_controller, dummy_plugin):
        assert list(plugin_controller.iter_plugins()) == [dummy_plugin]

    def test_emit(self, plugin_controller, dummy_plugin, extra_flags):
        rv = plugin_controller.emit("test-event")
        assert rv == {dummy_plugin.id: "test-event return value"}
        (call,) = dummy_plugin.calls
        assert call["extra_flags"] == extra_flags

    def test_emit_with_kwargs(self, plugin_controller, dummy_plugin):
        rv = plugin_controller.emit("another-test-event", another_arg="x")
        assert rv == {dummy_plugin.id: "another-test-event return value"}
        (call,) = dummy_plugin.calls
        assert call["kwargs"] == {"another_arg": "x"}

    def test_emit_deprecation_warning(self, plugin_controller, dummy_plugin):
        with pytest.deprecated_call():
            rv = plugin_controller.emit("legacy-event")
        assert rv == {dummy_plugin.id: "legacy-event return value"}

    def test_emit_is_not_confused_by_type_error(self, plugin_controller, dummy_plugin):
        # Excercises https://github.com/lektor/lektor/issues/1085
        with pytest.raises(TypeError):
            plugin_controller.emit("one-type-error")
        rv = plugin_controller.emit("one-type-error")
        assert rv == {dummy_plugin.id: "one-type-error return value"}


@pytest.mark.usefixtures("dummy_plugin_distribution")
def test_cli_integration(project, cli_runner, monkeypatch):
    """Check that plugin hooks receive extra_flags from command line."""
    # chdir to tests/demo-project
    monkeypatch.chdir(project.tree)
    result = cli_runner.invoke(
        cli, ["clean", "--yes", "-f", "flag1", "--extra-flag", "flag2:value2"]
    )
    print(result.output)
    assert result.exit_code == 0
    assert {"before-prune", "after-prune", "setup-env"}.issubset(
        call["event"] for call in DummyPlugin.calls
    )
    for call in DummyPlugin.calls:
        assert call["extra_flags"] == {"flag1": "flag1", "flag2": "value2"}



import datetime
import re
import sys
import weakref
from html import unescape
from pathlib import Path

import pytest

import lektor.context
from lektor.db import Pad
from lektor.environment import Environment


@pytest.fixture
def scratch_project_data(scratch_project_data):
    # Add a sub-page to the scratch project
    data = {"_model": "page", "title": "Subpage", "body": "Subpage body"}
    subpage_lr = scratch_project_data / "content/sub-page/contents.lr"
    subpage_lr.parent.mkdir()
    subpage_lr.write_text("".join(lektor.metaformat.serialize(data.items())))

    testbag_ini = scratch_project_data / "databags/testbag.ini"
    testbag_ini.parent.mkdir()
    testbag_ini.write_text("foo = bar")

    return scratch_project_data


@pytest.fixture
def compile_template(scratch_env):
    def compile_template(source, name="tmpl.html"):
        Path(scratch_env.root_path, "templates", name).write_text(
            source, encoding="utf-8"
        )
        return scratch_env.jinja_env.get_template(name)

    return compile_template


@pytest.fixture
def source_path():
    return "/"


@pytest.fixture
def bogus_context(scratch_pad, source_path):
    # Construct a Context that has a source, without going through all
    # all the steps necessary to construct an Artifact.
    with lektor.context.Context(pad=scratch_pad) as ctx:
        if source_path is not None:
            ctx.source = scratch_pad.get(source_path)
        yield


def test_jinja2_feature_autoescape(compile_template):
    tmpl = compile_template("{{ value }}", "tmpl.html")
    rendered = tmpl.render(value="<tag>")
    assert unescape(rendered) == "<tag>"
    assert "<" not in rendered


def test_jinja2_feature_with(compile_template):
    tmpl = compile_template("{% with x = 'good' %}{{ x }}{% endwith %}")
    assert tmpl.render() == "good"


def test_jinja2_feature_do(compile_template):
    tmpl = compile_template(
        "{% set x = ['a'] %}{% do x.append('b') %}{{ x|join('-') }}"
    )
    assert tmpl.render() == "a-b"


@pytest.mark.parametrize("source_path", [None, "/"])
@pytest.mark.usefixtures("bogus_context")
def test_jinja2_markdown_filter(compile_template):
    tmpl = compile_template("{{ '**word**' | markdown }}")
    assert "<strong>word</strong>" in tmpl.render()


@pytest.mark.usefixtures("bogus_context")
def test_jinja2_markdown_filter_resolve_links(compile_template):
    tmpl = compile_template(
        "{{ '[subpage](sub-page)' | markdown(resolve_links='always') }}"
    )
    assert re.search(r"<a.*\bhref=(['\"])sub-page/\1.*>subpage</a>", tmpl.render())


@pytest.mark.parametrize(
    "source_path, resolve_links",
    [
        (None, "if-possible"),
        (None, "never"),
        ("/", "never"),
    ],
)
@pytest.mark.usefixtures("bogus_context")
def test_jinja2_markdown_filter_noresolve_links(compile_template, resolve_links):
    tmpl = compile_template(
        f"{{{{ '[subpage](sub-page)' | markdown(resolve_links={resolve_links!r}) }}}}"
    )
    assert re.search(r"<a.*\bhref=(['\"])sub-page\1.*>subpage</a>", tmpl.render())


@pytest.mark.parametrize("source_path", [None])
@pytest.mark.usefixtures("bogus_context")
def test_jinja2_markdown_filter_resolve_raises_if_no_source_obj(compile_template):
    tmpl = compile_template(
        "{{ '[subpage](sub-page)' | markdown(resolve_links='always') }}"
    )
    with pytest.raises(RuntimeError) as exc_info:
        tmpl.render()
    assert re.search(r"\bsource object\b.*\brequired\b", str(exc_info.value))


@pytest.mark.skipif(
    not hasattr(sys, "getrefcount"), reason="interpreter does not support ref counting"
)
def test_no_reference_cycle_in_environment(project):
    ref = weakref.ref(project.make_env(load_plugins=False))
    # With ref counts (and no ref cycle), the environment should be immediately
    # garbage collected
    assert ref() is None


@pytest.fixture
def render_string(env):
    def render_string(s, **kwargs):
        template = env.jinja_env.from_string(s)
        return template.render(**kwargs)

    return render_string


def test_dateformat_filter(render_string):
    tmpl = "{{ dt | dateformat('yyyy-MM-dd') }}"
    dt = datetime.date(2001, 2, 3)
    assert render_string(tmpl, dt=dt) == "2001-02-03"


def test_datetimeformat_filter_not_inlined(pad):
    template = pad.env.jinja_env.from_string("{{ 1678749806 | datetimeformat }}")
    en_date = template.render()
    with lektor.context.Context(pad=pad) as ctx:
        ctx.source = pad.get("/", alt="de")
        de_date = template.render()
    assert en_date != de_date


def test_datetimeformat_filter(render_string):
    tmpl = "{{ dt | datetimeformat('yyyy-MM-ddTHH:mm') }}"
    dt = datetime.datetime(2001, 2, 3, 4, 5, 6)
    assert render_string(tmpl, dt=dt) == "2001-02-03T04:05"


def test_timeformat_filter(render_string):
    tmpl = "{{ dt | datetimeformat('HH:mm') }}"
    dt = datetime.time(1, 2, 3)
    assert render_string(tmpl, dt=dt) == "01:02"


@pytest.fixture(params=["dateformat", "datetimeformat", "timeformat"])
def dates_filter(request: pytest.FixtureRequest) -> str:
    return request.param


def test_dates_format_filter_handles_undefined(
    env: Environment, dates_filter: str
) -> None:
    template = env.jinja_env.from_string(f"{{{{ undefined | {dates_filter} }}}}")
    assert template.render() == ""


def test_dates_format_filter_raises_type_error_on_bad_arg(
    env: Environment, dates_filter: str
) -> None:
    template = env.jinja_env.from_string(f"{{{{ obj | {dates_filter} }}}}")
    with pytest.raises(TypeError, match="unexpected exception"):
        template.render(obj=object())


def test_dates_format_filter_raises_type_error_on_bad_format(
    env: Environment, dates_filter: str
) -> None:
    template = env.jinja_env.from_string(f"{{{{ now | {dates_filter}(42) }}}}")
    with pytest.raises(TypeError, match="should be a str"):
        template.render(now=datetime.datetime.now())


@pytest.mark.parametrize("arg", ["locale", "tzinfo"])
def test_dates_format_filter_raises_type_error_on_bad_kwarg(
    env: Environment, dates_filter: str, arg: str
) -> None:
    template = env.jinja_env.from_string(f"{{{{ now | {dates_filter}({arg}=42) }}}}")
    with pytest.raises(TypeError):
        template.render(now=datetime.datetime.now())


def test_bag_gets_site_from_jinja_context(
    scratch_env: Environment, scratch_pad: Pad
) -> None:
    template = scratch_env.jinja_env.from_string("{{ bag('testbag.foo') }}")
    assert template.render(site=scratch_pad) == "bar"


def test_bag_gets_site_from_lektor_context(
    scratch_env: Environment, scratch_pad: Pad
) -> None:
    template = scratch_env.jinja_env.from_string("{{ bag('testbag.foo') }}")
    with lektor.context.Context(pad=scratch_pad):
        assert template.render() == "bar"



"""Test that all modules/packages in the lektor tree are importable in any order

Here we import each module by itself, one at a time, each in a new
python interpreter.

"""

import pkgutil
import sys
from subprocess import run

import pytest

import lektor
from lektor.markdown import MISTUNE_VERSION


ignored = set()

# Do not check importability of module for the non-installed version of mistune
if MISTUNE_VERSION.startswith("2"):
    ignored.add("lektor.markdown.mistune0")
else:
    ignored.add("lektor.markdown.mistune2")


def iter_lektor_modules():
    for module in pkgutil.walk_packages(lektor.__path__, f"{lektor.__name__}."):
        if module.name not in ignored:
            yield module.name


@pytest.fixture(params=iter_lektor_modules())
def module(request):
    return request.param


@pytest.mark.slowtest
def test_import(module):
    python = sys.executable
    assert run([python, "-c", f"import {module}"], check=False).returncode == 0



from __future__ import annotations

import importlib.metadata
import tempfile
import urllib.parse
from typing import Any
from urllib.parse import urlsplit
from warnings import warn

from werkzeug import urls as werkzeug_urls
from werkzeug.datastructures import MultiDict

from lektor.utils import DeprecatedWarning


__all__ = ["werkzeug_urls_URL"]


_DEPRECATED_ATTRS = {
    "TemporaryDirectory": tempfile.TemporaryDirectory,
    "importlib_metadata": importlib.metadata,
}


def __getattr__(name):
    try:
        value = _DEPRECATED_ATTRS.get(name)
    except KeyError:
        # pylint: disable=raise-missing-from
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}") from None

    if hasattr(value, "__module__"):
        replacement = f"{value.__module__}.{value.__name__}"
    else:
        replacement = f"{value.__name__}"
    warn(
        DeprecatedWarning(
            name=f"lektor.compat.{name}",
            reason=f"use {replacement} instead",
            version="3.4.0",
        ),
        stacklevel=2,
    )
    return value


class _CompatURL(urllib.parse.SplitResult):
    """This is a replacement for ``werkzeug.urls.URL``.

    Here we implement those attributes and methods of ``URL`` which are
    likely to be used by existing Lektor publishing plugins.

    Currently unreimplemented here are the ``encode_netloc``, ``decode_netloc``,
    ``get_file_location``, and ``encode`` methods of ``werkzeug.urls.URL``.

    NB: Use of this class is deprecated. DO NOT USE THIS IN NEW CODE!

    """

    def __str__(self) -> str:
        return self.geturl()

    def replace(self, **kwargs: Any) -> _CompatURL:
        return self._replace(**kwargs)

    @property
    def host(self) -> str | None:
        return self.hostname

    @property
    def ascii_host(self) -> str | None:
        host = self.hostname
        if host is None:
            return None
        try:
            return host.encode("idna").decode("ascii")
        except UnicodeError:
            return host

    @property
    def auth(self) -> str | None:
        auth, _, _ = self.netloc.rpartition("@")
        return auth if auth != "" else None

    @property
    def username(self) -> str | None:
        username = super().username
        if username is None:
            return None
        return _unquote_legacy(username)

    @property
    def raw_username(self) -> str | None:
        return super().username

    @property
    def password(self) -> str | None:
        password = super().password
        if password is None:
            return None
        return _unquote_legacy(password)

    @property
    def raw_password(self) -> str | None:
        return super().password

    def decode_query(
        self,
        charset: str = "utf-8",
        include_empty: bool = True,
        errors: str = "replace",
        # parse_qsl does not support the separator parameter in python < 3.7.10.
        # separator: str = "&",
    ) -> MultiDict:
        return MultiDict(
            urllib.parse.parse_qsl(
                self.query,
                keep_blank_values=include_empty,
                encoding=charset,
                errors=errors,
                # separator=separator,
            )
        )

    def join(
        self, url: str | tuple[str, str, str, str, str], allow_fragments: bool = True
    ) -> _CompatURL:
        if isinstance(url, tuple):
            url = urllib.parse.urlunsplit(url)
        joined = urllib.parse.urljoin(self.geturl(), url, allow_fragments)
        return _CompatURL._make(urlsplit(joined))

    def to_url(self) -> str:
        return self.geturl()

    def to_uri_tuple(self) -> _CompatURL:
        return _CompatURL._make(urlsplit(werkzeug_urls.iri_to_uri(self.geturl())))

    def to_iri_tuple(self) -> _CompatURL:
        return _CompatURL._make(urlsplit(werkzeug_urls.uri_to_iri(self.geturl())))


def _unquote_legacy(value: str) -> str:
    try:
        return urllib.parse.unquote(value, "utf-8", "strict")
    except UnicodeError:
        return urllib.parse.unquote(value, "latin1")


# Provide a replacement for the deprecated werkzeug.urls.URL class
#
# NB: Do not use this in new code!
#
# We only use this in lektor.publishers in order to provide some backward
# compatibility for custom publishers from existing Lektor plugins.
# At such point as we decide that backward-compatibility is no longer
# needed, will be deleted.
#
werkzeug_urls_URL = getattr(werkzeug_urls, "URL", _CompatURL)



from __future__ import annotations

import codecs
import io
import json
import os
import posixpath
import re
import subprocess
import sys
import tempfile
import threading
import unicodedata
import urllib.parse
import uuid
import warnings
from collections.abc import Callable
from collections.abc import Hashable
from collections.abc import Iterable
from collections.abc import Iterator
from contextlib import contextmanager
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime
from functools import cache
from functools import wraps
from pathlib import Path
from pathlib import PurePosixPath
from typing import Any
from typing import ClassVar
from typing import IO
from typing import Literal
from typing import overload
from typing import TYPE_CHECKING
from typing import TypeVar

from jinja2 import is_undefined
from markupsafe import Markup
from slugify import slugify as _slugify
from werkzeug.http import http_date
from werkzeug.urls import iri_to_uri
from werkzeug.urls import uri_to_iri


if TYPE_CHECKING:
    from _typeshed import StrPath


is_windows = os.name == "nt"

_slash_escape = "\\/" not in json.dumps("/")

_last_num_re = re.compile(r"^(.*)(\d+)(.*?)$")
_list_marker = object()
_value_marker = object()

# Figure out our fs encoding, if it's ascii we upgrade to utf-8
fs_enc = sys.getfilesystemencoding()
try:
    if codecs.lookup(fs_enc).name == "ascii":
        fs_enc = "utf-8"
except LookupError:
    pass


def split_virtual_path(path):
    if "@" in path:
        return path.split("@", 1)
    return path, None


def _norm_join(a, b):
    return posixpath.normpath(posixpath.join(a, b))


def join_path(a, b):
    """Join two DB-paths.

    It is assumed that both paths are already normalized in that
    neither contains an extra "." or ".." components, double-slashes,
    etc.
    """
    # NB: This function is really only during URL resolution.  The only
    # place that references it is lektor.source.SourceObject._resolve_url.

    if posixpath.isabs(b):
        return b

    a_p, a_v = split_virtual_path(a)
    b_p, b_v = split_virtual_path(b)

    # Special case: paginations are considered special virtual paths
    # where the parent is the actual parent of the page.  This however
    # is explicitly not done if the path we join with refers to the
    # current path (empty string or dot).
    if b_p not in ("", ".") and a_v and a_v.isdigit():
        a_v = None

    # New path has a virtual path, add that to it.
    if b_v:
        rv = _norm_join(a_p, b_p) + "@" + b_v
    elif a_v:
        rv = a_p + "@" + _norm_join(a_v, b_p)
    else:
        rv = _norm_join(a_p, b_p)
    if rv[-2:] == "@.":
        rv = rv[:-2]
    return rv


def cleanup_path(path):
    # NB: POSIX allows for two leading slashes in a pathname, so we have to
    # deal with the possiblity of leading double-slash ourself.
    return posixpath.normpath("/" + path.lstrip("/"))


def cleanup_url_path(url_path):
    """Clean up a URL path.

    This strips any query, and/or fragment that may be present in the
    input path.

    Raises ValueError if the path contains a _scheme_
    which is neither ``http`` nor ``https``, or a _netloc_.
    """
    scheme, netloc, path, _, _ = urllib.parse.urlsplit(url_path, scheme="http")
    if scheme not in ("http", "https"):
        raise ValueError(f"Invalid scheme: {url_path!r}")
    if netloc:
        raise ValueError(f"Invalid netloc: {url_path!r}")

    # NB: POSIX allows for two leading slashes in a pathname, so we have to
    # deal with the possiblity of leading double-slash ourself.
    return posixpath.normpath("/" + path.lstrip("/"))


def parse_path(path):
    x = cleanup_path(path).strip("/").split("/")
    if x == [""]:
        return []
    return x


def is_path_child_of(a, b, strict=True):
    a_p, a_v = split_virtual_path(a)
    b_p, b_v = split_virtual_path(b)
    a_p = parse_path(a_p)
    b_p = parse_path(b_p)
    a_v = parse_path(a_v or "")
    b_v = parse_path(b_v or "")

    if not strict and a_p == b_p and a_v == b_v:
        return True
    if not a_v and b_v:
        return False
    if a_p == b_p and a_v[: len(b_v)] == b_v and len(a_v) > len(b_v):
        return True
    return a_p[: len(b_p)] == b_p and len(a_p) > len(b_p)


def untrusted_to_os_path(path):
    if not isinstance(path, str):
        path = path.decode(fs_enc, "replace")
    clean_path = cleanup_path(path)
    assert clean_path.startswith("/")
    return clean_path[1:].replace("/", os.path.sep)


def is_path(path):
    return os.path.sep in path or (os.path.altsep and os.path.altsep in path)


def magic_split_ext(filename, ext_check=True):
    """Splits a filename into base and extension.  If ext check is enabled
    (which is the default) then it verifies the extension is at least
    reasonable.
    """

    def bad_ext(ext):
        if not ext_check:
            return False
        if not ext or ext.split() != [ext] or ext.strip() != ext:
            return True
        return False

    parts = filename.rsplit(".", 2)
    if len(parts) == 1:
        return parts[0], ""
    if len(parts) == 2 and not parts[0]:
        return "." + parts[1], ""
    if len(parts) == 3 and len(parts[1]) < 5:
        ext = ".".join(parts[1:])
        if not bad_ext(ext):
            return parts[0], ext
    ext = parts[-1]
    if bad_ext(ext):
        return filename, ""
    basename = ".".join(parts[:-1])
    return basename, ext


def iter_dotted_path_prefixes(dotted_path):
    pieces = dotted_path.split(".")
    if len(pieces) == 1:
        yield dotted_path, None
    else:
        for x in range(1, len(pieces)):
            yield ".".join(pieces[:x]), ".".join(pieces[x:])


def resolve_dotted_value(obj, dotted_path):
    node = obj
    for key in dotted_path.split("."):
        if isinstance(node, dict):
            new_node = node.get(key)
            if new_node is None and key.isdigit():
                new_node = node.get(int(key))
        elif isinstance(node, list):
            try:
                new_node = node[int(key)]
            except (ValueError, TypeError, IndexError):
                new_node = None
        else:
            new_node = None
        node = new_node
        if node is None:
            break
    return node


def decode_flat_data(itemiter, dict_cls=dict):
    def _split_key(name):
        result = name.split(".")
        for idx, part in enumerate(result):
            if part.isdigit():
                result[idx] = int(part)
        return result

    def _enter_container(container, key):
        if key not in container:
            return container.setdefault(key, dict_cls())
        return container[key]

    def _convert(container):
        if _value_marker in container:
            force_list = False
            values = container.pop(_value_marker)
            if container.pop(_list_marker, False):
                force_list = True
                values.extend(_convert(x[1]) for x in sorted(container.items()))
            if not force_list and len(values) == 1:
                values = values[0]

            if not container:
                return values
            return _convert(container)
        if container.pop(_list_marker, False):
            return [_convert(x[1]) for x in sorted(container.items())]
        return dict_cls((k, _convert(v)) for k, v in container.items())

    result = dict_cls()

    for key, value in itemiter:
        parts = _split_key(key)
        if not parts:
            continue
        container = result
        for part in parts:
            last_container = container
            container = _enter_container(container, part)
            last_container[_list_marker] = isinstance(part, int)
        container[_value_marker] = [value]

    return _convert(result)


def merge(a, b):
    """Merges two values together."""
    if b is None and a is not None:
        return a
    if a is None:
        return b
    if isinstance(a, list) and isinstance(b, list):
        for idx, (item_1, item_2) in enumerate(zip(a, b, strict=False)):
            a[idx] = merge(item_1, item_2)
    if isinstance(a, dict) and isinstance(b, dict):
        for key, value in b.items():
            a[key] = merge(a.get(key), value)
        return a
    return a


def slugify(text):
    """
    A wrapper around python-slugify which preserves file extensions
    and forward slashes.
    """

    parts = text.split("/")
    parts[-1], ext = magic_split_ext(parts[-1])

    out = "/".join(_slugify(part) for part in parts)

    if ext:
        return out + "." + ext
    return out


def secure_filename(filename, fallback_name="file"):
    base = filename.replace("/", " ").replace("\\", " ")
    basename, ext = magic_split_ext(base)
    rv = slugify(basename).lstrip(".")
    if not rv:
        rv = fallback_name
    if ext:
        return rv + "." + ext
    return rv


def increment_filename(filename):
    directory, filename = os.path.split(filename)
    basename, ext = magic_split_ext(filename, ext_check=False)

    match = _last_num_re.match(basename)
    if match is not None:
        rv = match.group(1) + str(int(match.group(2)) + 1) + match.group(3)
    else:
        rv = basename + "2"

    if ext:
        rv += "." + ext
    if directory:
        return os.path.join(directory, rv)
    return rv


@cache
def locate_executable(exe_file, cwd=None, include_bundle_path=True):
    """Locates an executable in the search path."""
    choices = [exe_file]
    resolve = True

    # If it's already a path, we don't resolve.
    if os.path.sep in exe_file or (os.path.altsep and os.path.altsep in exe_file):
        resolve = False

    extensions = os.environ.get("PATHEXT", "").split(";")
    _, ext = os.path.splitext(exe_file)
    if (
        os.name != "nt"
        and "" not in extensions
        or any(ext.lower() == extension.lower() for extension in extensions)
    ):
        extensions.insert(0, "")

    if resolve:
        paths = os.environ.get("PATH", "").split(os.pathsep)
        choices = [os.path.join(path, exe_file) for path in paths]

    if os.name == "nt":
        choices.append(os.path.join((cwd or os.getcwd()), exe_file))

    try:
        for path in choices:
            for ext in extensions:
                if os.access(path + ext, os.X_OK):
                    return path + ext
        return None
    except OSError:
        return None


class JSONEncoder(json.JSONEncoder):
    def default(self, o):  # pylint: disable=method-hidden
        if is_undefined(o):
            return None
        if isinstance(o, datetime):
            return http_date(o)
        if isinstance(o, uuid.UUID):
            return str(o)
        if hasattr(o, "__html__"):
            return str(o.__html__())
        return json.JSONEncoder.default(self, o)


def htmlsafe_json_dump(obj, **kwargs):
    kwargs.setdefault("cls", JSONEncoder)
    rv = (
        json.dumps(obj, **kwargs)
        .replace("<", "\\u003c")
        .replace(">", "\\u003e")
        .replace("&", "\\u0026")
        .replace("'", "\\u0027")
    )
    if not _slash_escape:
        rv = rv.replace("\\/", "/")
    return rv


def tojson_filter(obj, **kwargs):
    return Markup(htmlsafe_json_dump(obj, **kwargs))


class Url(urllib.parse.SplitResult):
    """Make various parts of a URL accessible.

    This is the type of the values exposed by Lektor record fields of type "url".

    Since Lektor 3.4.0, this is essentially a `urllib.parse.SplitResult` as obtained by
    calling `urlsplit` on the URL normalized to an IRI.

    Generally, attributes such as ``netloc``, ``host``, ``path``, ``query``, and
    ``fragment`` return the IRI (internationalied) versions of those components.

    The URI (ASCII-encoded) version of the URL is available from the `ascii_url`
    attribute.

    NB: Changed in 3.4.0: The ``query`` attribute used to return the URI
    (ASCII-encoded) version of the query — I'm not sure why. Now it returns
    the IRI (internationalized) version of the query.

    """

    url: str

    def __new__(cls, value: str):
        # XXX: deprecate use of constructor so that eventually we can make its signature
        # match that of the SplitResult base class.
        warnings.warn(
            DeprecatedWarning(
                "Url",
                reason=(
                    "Direct construction of a Url instance is deprecated. "
                    "Use the Url.from_string classmethod instead."
                ),
                version="3.4.0",
            ),
            stacklevel=2,
        )
        return cls.from_string(value)

    @classmethod
    def from_string(cls, value: str) -> Url:
        """Construct instance from URL string.

        The input URL can be a URI (all ASCII) or an IRI (internationalized).
        """
        # The iri_to_uri operation is nominally idempotent — it can be passed either an
        # IRI or a URI (or something inbetween) and will return a URI.  So to fully
        # normalize input which can be either an IRI or a URI, first convert to URI,
        # then to IRI.
        iri = uri_to_iri(iri_to_uri(value))
        obj = cls._make(urllib.parse.urlsplit(iri))
        obj.url = value
        return obj

    def __str__(self) -> str:
        """The original un-normalized URL string."""
        return self.url

    @property
    def ascii_url(self) -> str:
        """The URL encoded to an all-ASCII URI."""
        return iri_to_uri(self.geturl())

    @property
    def ascii_host(self) -> str | None:
        """The hostname part of the URL IDNA-encoded to ASCII."""
        return urllib.parse.urlsplit(self.ascii_url).hostname

    @property
    def host(self) -> str | None:
        """The IRI (internationalized) version of the hostname.

        This attribute is provided for backwards-compatibility.  New code should use the
        ``hostname`` attribute instead.
        """
        return self.hostname

    @property
    def anchor(self) -> str:
        """The IRI (internationalized) version of the "anchor" part of the URL.

        This attribute is provided for backwards-compatibility.  New code should use the
        ``fragment`` attribute instead.
        """
        return self.fragment


def is_unsafe_to_delete(path, base):
    a = os.path.abspath(path)
    b = os.path.abspath(base)
    diff = os.path.relpath(a, b)
    first = diff.split(os.path.sep, maxsplit=1)[0]
    return first in (os.path.curdir, os.path.pardir)


def prune_file_and_folder(name, base):
    if is_unsafe_to_delete(name, base):
        return False
    try:
        os.remove(name)
    except OSError:
        try:
            os.rmdir(name)
        except OSError:
            return False
    head, tail = os.path.split(name)
    if not tail:
        head, tail = os.path.split(head)
    while head and tail:
        try:
            if is_unsafe_to_delete(head, base):
                return False
            os.rmdir(head)
        except OSError:
            break
        head, tail = os.path.split(head)
    return True


def sort_normalize_string(s):
    return unicodedata.normalize("NFD", str(s).lower().strip())


def get_dependent_url(url_path, suffix, ext=None):
    url_directory, url_filename = posixpath.split(url_path)
    url_base, url_ext = posixpath.splitext(url_filename)
    if ext is None:
        ext = url_ext
    return posixpath.join(url_directory, url_base + "@" + suffix + ext)


# These are the only modes we really support
_AtomicOpenTextMode = Literal["w", "wt", "tw", "r", "rt", "tr"]
_AtomicOpenBinaryModeWriting = Literal["wb", "bw"]
_AtomicOpenBinaryModeReading = Literal["rb", "br"]
_AtomicOpenMode = (
    _AtomicOpenTextMode | _AtomicOpenBinaryModeWriting | _AtomicOpenBinaryModeReading
)


@overload
@contextmanager
def atomic_open(
    filename: StrPath,
    mode: _AtomicOpenTextMode = "r",
    encoding: str | None = None,
) -> Iterator[io.TextIOWrapper]: ...


@overload
@contextmanager
def atomic_open(
    filename: StrPath,
    mode: _AtomicOpenBinaryModeWriting,
    encoding: None = None,
) -> Iterator[io.BufferedWriter]: ...


@overload
@contextmanager
def atomic_open(
    filename: StrPath,
    mode: _AtomicOpenBinaryModeReading,
    encoding: None = None,
) -> Iterator[io.BufferedReader]: ...


@contextmanager
def atomic_open(
    filename: StrPath, mode: _AtomicOpenMode = "r", encoding: str | None = None
) -> Iterator[IO[Any]]:
    """Open a file for atomic update.

    Perform an "all-or-nothing" write to a file.

    This opens a temporary file to receive writes. It is meant to be used as a context
    manager. When the context is exited normally, the temporary file is closed then
    atomically renamed to the target file name.

    If an exception is thrown during this process the temporary file is silently
    deleted.

    """
    if any(c in mode for c in "ax+"):
        raise ValueError(f"unsupported open mode: {mode}")

    if "r" in mode:
        with open(filename, mode=mode, encoding=encoding) as fp:
            yield fp
        return

    fd, tmp_filename = create_temp(
        prefix=".__atomic-write",
        dir=Path(filename).parent,
        text="b" not in mode,
    )
    try:
        with open(fd, mode, encoding=encoding) as fp:
            yield fp
        os.replace(tmp_filename, filename)
    except Exception:
        with suppress(OSError):
            os.remove(tmp_filename)
        raise


def create_temp(
    suffix: str = "",
    prefix: str = "tmp",
    dir: StrPath | None = None,
    text: bool = False,
    mode: int = 0o666,
) -> tuple[int, str | bytes]:
    """Create and return a unique temporary file.

    The return value is a pair (fd, name) where fd is the file descriptor returned by
    os.open, and name is the filename.

    This works very much like `tempfile.mkstemp`, except that it allows more control
    over the mode (access permissions) of the created file.

    The access permissions of the created file is determined by the value of 'mode'
    (which defaults to 0o666) combined with any umask or default ACL that may be in
    place.

    """
    if (tmp_dir := dir) is None:
        tmp_dir = os.getcwd()

    flags = os.O_RDWR | os.O_CREAT | os.O_EXCL | getattr(os, "O_NOFOLLOW", 0)
    if not text:
        flags |= getattr(os, "O_BINARY", 0)

    for _ in range(32000):
        filename = tempfile.mktemp(suffix=suffix, prefix=prefix, dir=tmp_dir)
        try:
            fd = os.open(filename, flags, mode)
        except FileExistsError:
            continue
        # NB: Under Windows, PermissionError can be raised by os.open if a directory
        # with the chosen name already exists. Since mktemp doesn't return names to
        # existing files/dirs, the likelihood of this happening is slim. For now, we
        # don't check for it here.
        return fd, filename

    raise AssertionError("Unable to find temporary file name")


def portable_popen(cmd, *args, **kwargs):
    """A portable version of subprocess.Popen that automatically locates
    executables before invoking them.  This also looks for executables
    in the bundle bin.
    """
    if cmd[0] is None:
        raise RuntimeError("No executable specified")
    exe = locate_executable(cmd[0], kwargs.get("cwd"))
    if exe is None:
        raise RuntimeError(f'Could not locate executable "{cmd[0]}"')

    if isinstance(exe, str) and sys.platform != "win32":
        exe = exe.encode(sys.getfilesystemencoding())
    cmd[0] = exe
    return subprocess.Popen(cmd, *args, **kwargs)


def is_valid_id(value):
    if value == "":
        return True
    return (
        "/" not in value
        and value.strip() == value
        and value.split() == [value]
        and not value.startswith(".")
    )


def secure_url(url: str) -> str:
    parts = urllib.parse.urlsplit(url)
    if parts.password is not None:
        _, _, host_port = parts.netloc.rpartition("@")
        parts = parts._replace(netloc=f"{parts.username}@{host_port}")
    return parts.geturl()


def bool_from_string(val, default=None):
    if val in (True, False, 1, 0):
        return bool(val)
    if isinstance(val, str):
        val = val.lower()
        if val in ("true", "yes", "1"):
            return True
        if val in ("false", "no", "0"):
            return False
    return default


def make_relative_url(source, target):
    """
    Returns the relative path (url) needed to navigate
    from `source` to `target`.
    """

    # WARNING: this logic makes some unwarranted assumptions about
    # what is a directory and what isn't. Ideally, this function
    # would be aware of the actual filesystem.
    s_is_dir = source.endswith("/")
    t_is_dir = target.endswith("/")

    source = PurePosixPath(posixpath.normpath(source))
    target = PurePosixPath(posixpath.normpath(target))

    if not s_is_dir:
        source = source.parent

    relpath = str(get_relative_path(source, target))
    if t_is_dir:
        relpath += "/"

    return relpath


def get_relative_path(source, target):
    """
    Returns the relative path needed to navigate from `source` to `target`.

    get_relative_path(source: PurePosixPath,
                      target: PurePosixPath) -> PurePosixPath
    """

    if not source.is_absolute() and target.is_absolute():
        raise ValueError("Cannot navigate from a relative path to an absolute one")

    if source.is_absolute() and not target.is_absolute():
        # nothing to do
        return target

    if source.is_absolute() and target.is_absolute():
        # convert them to relative paths to simplify the logic
        source = source.relative_to("/")
        target = target.relative_to("/")

    # is the source an ancestor of the target?
    try:
        return target.relative_to(source)
    except ValueError:
        pass

    # even if it isn't, one of the source's ancestors might be
    # (and if not, the root will be the common ancestor)
    distance = PurePosixPath(".")
    for ancestor in source.parents:
        distance /= ".."

        try:
            relpath = target.relative_to(ancestor)
        except ValueError:
            continue
        else:
            # prepend the distance to the common ancestor
            return distance / relpath
    # We should never get here.  (The last ancestor in source.parents will
    # be '.' — target.relative_to('.') will always succeed.)
    raise AssertionError("This should not happen")


def deg_to_dms(deg):
    d = int(deg)
    md = abs(deg - d) * 60
    m = int(md)
    sd = (md - m) * 60
    return (d, m, sd)


def format_lat_long(lat=None, long=None, secs=True):
    def _format(value, sign):
        d, m, sd = deg_to_dms(value)
        return f"{abs(d)}° {abs(m)}′ {secs and f'{abs(sd)}″ ' or ''}{sign[d < 0]}"

    rv = []
    if lat is not None:
        rv.append(_format(lat, "NS"))
    if long is not None:
        rv.append(_format(long, "EW"))
    return ", ".join(rv)


def get_cache_dir():
    if is_windows:
        folder = os.environ.get("LOCALAPPDATA")
        if folder is None:
            folder = os.environ.get("APPDATA")
            if folder is None:
                folder = os.path.expanduser("~")
        return os.path.join(folder, "Lektor", "Cache")
    if sys.platform == "darwin":
        return os.path.join(os.path.expanduser("~/Library/Caches/Lektor"))
    return os.path.join(
        os.environ.get("XDG_CACHE_HOME", os.path.expanduser("~/.cache")), "lektor"
    )


class URLBuilder:
    def __init__(self):
        self.items = []

    def append(self, item):
        if item is None:
            return
        item = str(item).strip("/")
        if item:
            self.items.append(item)

    def get_url(self, trailing_slash=None):
        url = "/" + "/".join(self.items)
        if trailing_slash is not None and not trailing_slash:
            return url
        if url == "/":
            return url
        if trailing_slash is None:
            _, last = url.rsplit("/", 1)
            if "." in last:
                # Assuming the url points to a file
                return url
        return url + "/"


def build_url(iterable, trailing_slash=None):
    # NB: While this function is not used by Lektor itself, it is used
    # by a number of plugins including: lektor-atom,
    # lektor-gemini-capsule, lektor-index-pages, and lektor-tags.
    builder = URLBuilder()
    for item in iterable:
        builder.append(item)
    return builder.get_url(trailing_slash=trailing_slash)


def comma_delimited(s):
    """Split a comma-delimited string."""
    for part in s.split(","):
        stripped = part.strip()
        if stripped:
            yield stripped


def process_extra_flags(flags):
    if isinstance(flags, dict):
        return flags
    rv = {}
    for flag in flags or ():
        if ":" in flag:
            k, v = flag.split(":", 1)
            rv[k] = v
        else:
            rv[flag] = flag
    return rv


_H = TypeVar("_H", bound=Hashable)


def unique_everseen(seq: Iterable[_H]) -> Iterable[_H]:
    """Filter out duplicates from iterable."""
    # This is a less general version of more_itertools.unique_everseen.
    # Should we need more general functionality, consider using that instead.
    seen = set()
    for val in seq:
        if val not in seen:
            seen.add(val)
            yield val


class RecursionCheck(threading.local):
    """A context manager that retains a count of how many times it's been entered.

    Example:

        >>> recursion_check = RecursionCheck()

        >>> with recursion_check:
        ...     assert recursion_check.level == 1
        ...     with recursion_check as recursion_level:
        ...         assert recursion_check.level == 2
        ...         print("depth", recursion_level)
        ...     assert recursion_check.level == 1
        ... assert recursion_check.level == 0
        depth 2
    """

    level = 0

    def __enter__(self) -> bool:
        self.level += 1
        return self.level

    def __exit__(self, _t, _v, _tb) -> None:
        self.level -= 1


class DeprecatedWarning(DeprecationWarning):
    """Warning category issued by our ``deprecated`` decorator."""

    def __init__(
        self,
        name: str,
        reason: str | None = None,
        version: str | None = None,
    ):
        self.name = name
        self.reason = reason
        self.version = version

    def __str__(self) -> str:
        message = f"{self.name!r} is deprecated"
        if self.reason:
            message += f" ({self.reason})"
        if self.version:
            message += f" since version {self.version}"
        return message


_F = TypeVar("_F", bound=Callable[..., Any])


@dataclass
class _Deprecate:
    """A decorator to mark callables as deprecated."""

    name: str | None = None
    reason: str | None = None
    version: str | None = None
    stacklevel: int = 1

    _recursion_check: ClassVar = RecursionCheck()

    def __call__(self, wrapped: _F) -> _F:
        if not callable(wrapped):
            raise TypeError("do not know how to deprecate {wrapped!r}")

        name = self.name or wrapped.__name__
        message = DeprecatedWarning(name, self.reason, self.version)

        @wraps(wrapped)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            with self._recursion_check as recursion_level:
                if recursion_level == 1:
                    warnings.warn(message, stacklevel=self.stacklevel + 1)
                return wrapped(*args, **kwargs)

        return wrapper  # type: ignore[return-value]


@overload
def deprecated(
    __wrapped: Callable[..., Any],
    *,
    name: str | None = ...,
    reason: str | None = ...,
    version: str | None = ...,
    stacklevel: int = ...,
) -> Callable[..., Any]: ...


@overload
def deprecated(
    __reason: str,
    *,
    name: str | None = ...,
    version: str | None = ...,
    stacklevel: int = ...,
) -> _Deprecate: ...


@overload
def deprecated(
    *,
    name: str | None = ...,
    reason: str | None = ...,
    version: str | None = ...,
    stacklevel: int = ...,
) -> _Deprecate: ...


def deprecated(*args: Any, **kwargs: Any) -> _F | _Deprecate:
    """A decorator to mark callables or descriptors as deprecated.

    This can be used to decorate functions, methods, classes, and descriptors.
    In particular, this decorator can be applied to instances of ``property``,
    ``functools.cached_property`` and ``werkzeug.utils.cached_property``.

    When the decorated object is called (or — in the case of a descriptor — accessed), a
    ``DeprecationWarning`` is issued.

    The warning message will include the name of the decorated object, and may include
    further information if provided from the ``reason`` and ``version`` arguments.

    The ``name`` argument may be used to specify an alternative name to use when
    generating the warning message. By default, the ``__name__`` attribute of the
    decorated object is used.

    The ``stacklevel`` argument controls which call in the call stack the warning
    is attributed to. The default value, ``stacklevel=1`` means the warning is
    reported for the immediate caller of the decorated object.  Higher values
    attribute the warning callers further back in the stack.

    """
    if len(args) > 1:
        raise TypeError("deprecated accepts a maximum of one positional parameter")

    wrapped: _F | None = None
    if args:
        if isinstance(args[0], str):
            kwargs.setdefault("reason", args[0])
        else:
            wrapped = args[0]

    deprecate = _Deprecate(**kwargs)
    if wrapped is not None:
        return deprecate(wrapped)
    return deprecate



import sys
from contextlib import suppress
from importlib import import_module
from importlib import metadata


# pylint: disable-next=wrong-import-order
from conftest import restore_import_state  # isort: skip


def test_restore_import_state_restores_meta_path():
    # Various modules (e.g. setuptools, importlib_metadata), when imported,
    # add their own finders to sys.meta_path.
    meta_path = sys.meta_path.copy()

    with restore_import_state(), suppress(ModuleNotFoundError):
        import_module("importlib_metadata")

    assert sys.meta_path == meta_path


def test_restore_import_state_restores_unneutered_PathFinder():
    # When importlib_metadata is imported, it neuters the stdlib
    # distribution find, and then adds its own finder to meta_path.
    #
    # This tests that restore_import_state manages to unneuter
    # this find.
    distributions_pre = [dist.metadata["name"] for dist in metadata.distributions()]

    with restore_import_state(), suppress(ModuleNotFoundError):
        import_module("importlib_metadata")

    distributions_post = [dist.metadata["name"] for dist in metadata.distributions()]
    assert len(distributions_pre) == len(distributions_post)
    assert set(distributions_pre) == set(distributions_post)



import sys
import warnings
from collections.abc import Hashable
from importlib import metadata
from typing import Any
from typing import TYPE_CHECKING
from weakref import ref as weakref

from markupsafe import Markup

from lektor.markdown.controller import ControllerCache
from lektor.markdown.controller import FieldOptions
from lektor.markdown.controller import MarkdownController
from lektor.markdown.controller import Meta
from lektor.markdown.controller import RenderResult
from lektor.sourceobj import SourceObject
from lektor.utils import deprecated
from lektor.utils import DeprecatedWarning


if TYPE_CHECKING:  # pragma: no cover
    from lektor.environment import Environment

controller_class: type[MarkdownController]


MISTUNE_VERSION = metadata.version("mistune")
if MISTUNE_VERSION.startswith("0."):
    from lektor.markdown.mistune0 import MarkdownController0 as controller_class
elif MISTUNE_VERSION.startswith("2."):
    from lektor.markdown.mistune2 import MarkdownController2 as controller_class
else:  # pragma: no cover
    raise ImportError("Unsupported version of mistune")


get_controller = ControllerCache(controller_class)


class Markdown:
    def __init__(
        self, source: str, record: SourceObject | None, field_options: FieldOptions
    ) -> None:
        self.source = source
        self.__record = weakref(record) if record is not None else None
        self.__field_options = field_options
        self.__cache: dict[Hashable, RenderResult] = {}

    def __bool__(self) -> bool:
        return bool(self.source)

    __nonzero__ = __bool__

    @property
    def record(self) -> SourceObject:
        ref = self.__record
        if ref is None:
            return None
        record = ref()
        if record is None:
            raise RuntimeError("Record has gone away")
        return record

    def __render(self) -> RenderResult:
        # When the markdown instance is attached to a cached object we
        # can end up in the situation where, e.g., the base_url has
        # changed from the time we were put into the cache to the time
        # where we got referenced by something elsewhere.  Since this
        # affects the processing of relative links, in that case we
        # need to re-process our markdown.
        controller = get_controller()
        key = controller.get_cache_key()
        result = self.__cache.get(key) if key is not None else None
        if result is None:
            result = controller.render(self.source, self.record, self.__field_options)
            if key is not None:
                self.__cache[key] = result
        return result

    @property
    def meta(self) -> Meta:
        return self.__render().meta

    @property
    def html(self) -> Markup:
        return Markup(self.__render().html)

    def __getitem__(self, name: str) -> Any:
        return self.meta[name]

    def __str__(self) -> str:
        return self.__render().html

    def __html__(self) -> Markup:
        return self.html


# Deprecated methods and attributes follow
#
# It is hoped that these are sufficient that most plugins that extend
# the markdown processing will continue to work (so long as mistune is
# pinned to 0.x).
_deprecated_moved_to_submodule = {"escape", "ImprovedRenderer", "MarkdownConfig"}


def __getattr__(name):
    """Access to ``escape``, ``ImprovedRenderer``, and ``MarkdownConfig``.

    These are imported from from either our .mistune0 or .mistune2 modules,
    as appropriate for the installed version of mistune.
    """
    if name not in _deprecated_moved_to_submodule:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    message = DeprecatedWarning(f"{__name__}.{name}", version="3.4.0")
    warnings.warn(message, stacklevel=2)
    mistune_module = sys.modules[controller_class.__module__]
    return getattr(mistune_module, name)


def __dir__():
    return sorted(set(globals().keys()) | _deprecated_moved_to_submodule)


@deprecated(version="3.4.0")
def make_markdown(env: "Environment") -> Any:  # (Environment) -> mistune.Markdown
    return get_controller(env).make_parser()


@deprecated(version="3.4.0")
def markdown_to_html(
    text: str, record: SourceObject, field_options: FieldOptions
) -> RenderResult:
    return get_controller().render(text, record, field_options)



"""Tests for the hatch build hooks in ../build_frontend.py."""

import os
import shutil
import sys
from importlib.util import module_from_spec
from importlib.util import spec_from_file_location
from pathlib import Path
from unittest.mock import Mock

import pytest


@pytest.fixture(scope="session")
def frontend_build_module():
    return import_module_from_file("build_frontend", "../build_frontend.py")


@pytest.fixture
def tmp_root(tmp_path):
    root = tmp_path / "root"
    lektor_admin = root / "lektor/admin"
    lektor_admin.mkdir(parents=True)
    return root


@pytest.fixture
def frontend_src(tmp_root):
    """Copy the frontend source to tmp_root"""
    top = Path(__file__).parent / ".."
    for path in "frontend", "lektor/translations":
        dst = tmp_root / path
        dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copytree(top / path, dst, ignore=shutil.ignore_patterns("node_modules"))


@pytest.fixture
def compiled_output(tmp_root):
    app_js = tmp_root / "lektor/admin/static/app.js"
    app_js.parent.mkdir(parents=True, exist_ok=True)
    app_js.touch()
    return app_js


@pytest.fixture
def frontend_build_hook(frontend_build_module, tmp_root, tmp_path):
    return frontend_build_module.FrontendBuildHook(
        root=tmp_root,
        config={},
        build_config=Mock(name="BuilderConfig"),
        metadata=Mock(name="ProjectMetadata"),
        directory=tmp_path / "dist",
        target_name="wheel",
    )


@pytest.mark.usefixtures("frontend_src")
def test_clean(frontend_build_hook, compiled_output):
    frontend_build_hook.clean(("standard",))
    assert not compiled_output.parent.exists()


@pytest.mark.usefixtures("frontend_src")
def test_clean_idempotent(frontend_build_hook, compiled_output):
    frontend_build_hook.clean(("standard",))
    assert not compiled_output.parent.exists()
    frontend_build_hook.clean(("standard",))
    assert not compiled_output.parent.exists()


def test_clean_skipped_if_no_frontend(frontend_build_hook, compiled_output):
    frontend_build_hook.clean(("standard",))
    assert compiled_output.parent.exists()


@pytest.mark.usefixtures("frontend_src")
def test_initialize_skips_build_if_output_exists(
    frontend_build_hook, compiled_output, mocker
):
    mocker.patch("shutil.which").return_value = None
    frontend_build_hook.initialize("standard", build_data={})
    assert compiled_output.exists()


@pytest.mark.usefixtures("frontend_src")
def test_initialize_aborts_if_no_npm(
    frontend_build_hook, frontend_build_module, monkeypatch
):
    monkeypatch.setitem(os.environ, "PATH", "")
    with pytest.raises(SystemExit) as exc_info:
        frontend_build_hook.initialize("standard", build_data={})
    assert exc_info.value.args[0] == 1


def test_initialize_aborts_if_no_source(frontend_build_hook, mocker):
    mocker.patch("shutil.which").return_value = "/bin/false"
    with pytest.raises(SystemExit) as exc_info:
        frontend_build_hook.initialize("standard", build_data={})
    assert exc_info.value.args[0] == 1


@pytest.mark.skipif(shutil.which("npm") is None, reason="npm not installed")
@pytest.mark.requiresinternet
@pytest.mark.slowtest
@pytest.mark.usefixtures("frontend_src")
def test_initialize_builds_frontend(frontend_build_hook, tmp_root):
    frontend_build_hook.initialize("standard", build_data={})
    app_js = tmp_root / "lektor/admin/static/app.js"
    assert app_js.stat().st_size > 1024


def import_module_from_file(module_name: str, path: str) -> None:
    """Import a module or package from a specific source (``.py``) file

    This bypasses the normal search of ``sys.path``, etc., directly
    importing the module from the specified python source file.

    The imported module is registered, as usual, in ``sys.modules``.

    If the path to the source file is relative, it is interpreted
    relative to the directory containing this script.

    """
    # Copied more-or-less verbatim from:
    # https://docs.python.org/3/library/importlib.html?highlight=import#importing-a-source-file-directly
    here = Path(__file__).parent
    spec = spec_from_file_location(module_name, here / path)
    if spec is None or spec.loader is None:
        raise ModuleNotFoundError(f"Can not find {module_name}")
    module = module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module



from __future__ import annotations

import hashlib
import os
import shutil
import site
import subprocess
import sys
import sysconfig
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sized
from pathlib import Path
from typing import TYPE_CHECKING
from venv import EnvBuilder

import click
import requests


if TYPE_CHECKING:
    from _typeshed import StrPath

    from lektor.environment import Environment  # circ dependency
else:
    StrPath = object


def _get_package_version_from_project(cfg, name):
    choices = (name.lower(), "lektor-" + name.lower())
    for pkg, version in cfg.section_as_dict("packages").items():
        if pkg.lower() in choices:
            return {"name": pkg, "version": version}
    return None


def add_package_to_project(project, req):
    """Given a package requirement this returns the information about this
    plugin.
    """
    if "@" in req:
        name, version = req.split("@", 1)
        version_hint = version
    else:
        name = req
        version = None
        version_hint = "latest release"

    cfg = project.open_config()
    info = _get_package_version_from_project(cfg, name)
    if info is not None:
        raise RuntimeError("The package was already added to the project.")

    for choice in name, "lektor-" + name:
        rv = requests.get(f"https://pypi.python.org/pypi/{choice}/json", timeout=10)
        if rv.status_code != 200:
            continue

        data = rv.json()
        canonical_name = data["info"]["name"]
        if version is None:
            version = data["info"]["version"]
        version_info = data["releases"].get(version)
        if version_info is None:
            raise RuntimeError(
                f"Latest requested version ({version_hint}) could not be found"
            )

        cfg[f"packages.{canonical_name}"] = version
        cfg.save()
        return {"name": canonical_name, "version": version}

    raise RuntimeError("The package could not be found on PyPI")


def remove_package_from_project(project, name):
    cfg = project.open_config()
    choices = (name.lower(), "lektor-" + name.lower())
    for pkg, version in cfg.section_as_dict("packages").items():
        if pkg.lower() in choices:
            del cfg[f"packages.{pkg}"]
            cfg.save()
            return {"name": pkg, "version": version}
    return None


if os.name == "nt":
    _default_venv_symlinks = False
else:
    _default_venv_symlinks = True


class VirtualEnv:
    """A helper for manipulating our private package cache virtual environment.

    Parameters:

    path — The path to the virtual environment to manage.  This can be an existing
    environment or not.

    """

    def __init__(self, path: StrPath):
        self.path = Path(path)

    def create(
        self,
        with_pip: bool = True,
        upgrade_deps: bool = True,
        symlinks: bool = _default_venv_symlinks,
    ) -> None:
        """(Re-)Create a new virtual environment.

        This will remove any existing virtual environment and create a new one.

        The parameters ``with_pip`` and ``upgrade_deps`` should probably be left at
        their default values in normal usage.  They are provided here mostly for use in
        tests. They work as described for ``venv.EnvBuilder`` from the standard library.
        (Though ``upgrade_deps`` is only supported by EnvBuilder`` in py39+, here we
        emulate it's behavior if running under older pythons.)

        """
        # Right now, by default, we always install and upgrade pip to
        # the latest available version.
        #
        # We could optimize by not installing (and not upgrading) pip if
        # the system pip is sufficient to our needs.
        #
        # Note that, e.g., pip>=21.3 is required to support PEP660 editable
        # installs.
        env_builder = EnvBuilder(
            clear=True,
            with_pip=with_pip,
            upgrade_deps=upgrade_deps,
            symlinks=symlinks,
        )
        env_builder.create(self.path)

    def addsitedir(self, sitedir: str) -> None:
        """Add an additional sitedir to sys.path for virtual environment.

        Packages installed in ``sitedir`` will be made available to any invocations
        of python running within the virtual environment.
        """
        with Path(self.site_packages, "_lektor.pth").open("a", encoding="utf-8") as fp:
            fp.write(f"import site; site.addsitedir({sitedir!r})\n")

    def run_pip_install(self, *args: str) -> None:
        """Run `pip install` in the virtual environment.

        ``Args`` are appended to the command line (following ``pip install``). They
        should specify how and which packages to install.

        """
        try:
            subprocess.run((self.executable, "-m", "pip", "install", *args), check=True)
        except subprocess.CalledProcessError as exc:
            raise RuntimeError("Failed to install dependency package.") from exc

    @property
    def site_packages(self) -> str:
        """The path to the virtual environments ``site-packages`` directory."""
        return self._get_path("purelib")

    @property
    def executable(self) -> str:
        """The path to the python interpreter for the virtual environment."""
        script_path = Path(self._get_path("scripts"))
        executable_name = Path(sys.executable).name
        return os.fspath(script_path / executable_name)

    def _get_path(self, name: str) -> str:
        vars = {"base": os.fspath(self.path)}
        return sysconfig.get_path(name, vars=vars)


class Requirements(Iterable[str], Sized):
    """Manage package requirements."""

    requirements: set[str]

    def __init__(self) -> None:
        self.requirements = set()

    def __len__(self) -> int:
        return len(self.requirements)

    def __iter__(self) -> Iterator[str]:
        """The requirements.

        These requirements are in the form of arguments that can be passed to ``pip
        install``.
        """
        return iter(self.requirements)

    def add_requirement(self, package: str, version: str | None = None) -> None:
        """Add a (remote) distribution to the requirements."""
        self.requirements.add(f"{package}=={version}" if version else f"{package}")

    def add_local_requirement(self, path: StrPath) -> None:
        """Add a local distribution source directory to the requirements.

        The distribution source at ``path`` (which should be a legacy `setup.py` or
        modern PEP660-compatible project) will be installed in editable mode.

        """
        srcdir = os.fspath(Path(path).resolve())
        self.requirements.add(f"--editable={srcdir}")

    _DIST_FILES = ("setup.py", "pyproject.toml")

    def add_local_requirements_from(self, packages_path: StrPath) -> None:
        """Add sub-directories of path that look like local distribution sources.

        Any direct sub-directories of ``packages_path`` which appear to be distribution
        source code will be added to the requirements in local (editable) mode.

        """
        try:
            for path in Path(packages_path).iterdir():
                if any(path.joinpath(fn).is_file() for fn in self._DIST_FILES):
                    self.add_local_requirement(path)
        except OSError:
            pass

    def hash(self) -> str:
        """Compute a hash of the requirement set."""
        hash = hashlib.sha1()
        for requirement in sorted(self.requirements):
            hash.update(requirement.encode("utf-8"))
            hash.update(b"\0")
        return hash.hexdigest()


def update_cache(
    venv_path: Path,
    remote_packages: dict[str, str],
    local_package_path: Path,
) -> None:
    """Ensure the package cache at venv_path is up-to-date.

    ``Remote_packages`` is a dictionary (mapping package names to required versions)
    that specifies remote packages (to be installed from PyPI).

    ``Local_package_page`` is a path to a directory whose sub-directories may contain
    local plugin source.  Any such source directories will be installed in "editable"
    mode.

    """
    requirements = Requirements()
    for package, version in remote_packages.items():
        requirements.add_requirement(package, version)
    requirements.add_local_requirements_from(local_package_path)

    if len(requirements) == 0:
        shutil.rmtree(venv_path, ignore_errors=True)
    else:
        hash_file = venv_path / "lektor-requirements-hash.txt"
        try:
            is_stale = hash_file.read_text().strip() != requirements.hash()
        except FileNotFoundError:
            is_stale = True

        if is_stale:
            venv = VirtualEnv(venv_path)
            venv.create()
            # Add our site-packages to venv's sys.path
            our_site_packages = sysconfig.get_path("purelib")
            venv.addsitedir(our_site_packages)

            venv.run_pip_install(*requirements)
            hash_file.write_text(f"{requirements.hash()}\n", encoding="ascii")


def load_packages(env: Environment, reinstall: bool = False) -> None:
    """Import all of our managed plugins into our ``sys.path``

    This first ensures that our private package cache is up-to-date, then
    adds it to ``sys.path``.

    After ``load_packages`` is called, the entry points defined in by
    plugins that we manage will be available.
    """
    if reinstall:
        click.echo("Force package cache refresh.")
        wipe_package_cache(env)

    config = env.load_config()
    venv_path = env.project.get_package_cache_path()
    update_cache(venv_path, config["PACKAGES"], Path(env.root_path, "packages"))
    site.addsitedir(VirtualEnv(venv_path).site_packages)


def wipe_package_cache(env: Environment) -> None:
    """Remove the entire package cache."""
    project = env.project
    # Remove the legacy flat package cache, too
    for cache_type in project.PackageCacheType:
        shutil.rmtree(project.get_package_cache_path(cache_type), ignore_errors=True)



from __future__ import annotations

from pathlib import Path

from ..utils import deprecated
from .exif import read_exif
from .image_info import get_image_info
from .thumbnail import compute_dimensions
from .thumbnail import make_image_thumbnail
from .thumbnail import Thumbnail
from .thumbnail import ThumbnailMode


__all__ = [
    "compute_dimensions",
    "get_image_info",
    "get_quality",
    "make_image_thumbnail",
    "read_exif",
    "Thumbnail",
    "ThumbnailMode",
]


@deprecated(version="3.4.0")
def get_quality(source_filename: str | Path) -> int:
    """Get the effective default thumbnail _quality_.

    This is the ImageMagick "quality" that is used, by default, when generating
    thumbnails.

    """
    if get_image_info(source_filename).format == "png":
        return 75
    return 85



import re
from datetime import date
from datetime import datetime

from babel.dates import get_timezone
from markupsafe import Markup

from lektor.constants import PRIMARY_ALT
from lektor.i18n import get_i18n_block
from lektor.types.base import Type
from lektor.utils import bool_from_string


class SingleInputType(Type):
    widget = "singleline-text"

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = Type.to_json(self, pad, record, alt)
        rv["addon_label_i18n"] = get_i18n_block(self.options, "addon_label") or None
        return rv


class StringType(SingleInputType):
    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing string")
        try:
            return raw.value.splitlines()[0].strip()
        except IndexError:
            return ""


class StringsType(Type):
    widget = "multiline-text"

    def value_from_raw(self, raw):
        return [x.strip() for x in (raw.value or "").splitlines()]


class TextType(Type):
    widget = "multiline-text"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing text")
        return raw.value


class HtmlType(Type):
    widget = "multiline-text"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing HTML")
        return Markup(raw.value)


class IntegerType(SingleInputType):
    widget = "integer"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing integer value")
        try:
            return int(raw.value.strip())
        except ValueError:
            try:
                return int(float(raw.value.strip()))
            except ValueError:
                return raw.bad_value("Not an integer")


class FloatType(SingleInputType):
    widget = "float"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing float value")
        try:
            return float(raw.value.strip())
        except ValueError:
            return raw.bad_value("Not an integer")


class BooleanType(Type):
    widget = "checkbox"

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = Type.to_json(self, pad, record, alt)
        rv["checkbox_label_i18n"] = get_i18n_block(self.options, "checkbox_label")
        return rv

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing boolean")
        val = bool_from_string(raw.value.strip().lower())
        if val is None:
            return raw.bad_value("Bad boolean value")
        return val


class DateType(SingleInputType):
    widget = "datepicker"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing date")
        try:
            return date(*map(int, raw.value.split("-")))
        except Exception:
            return raw.bad_value("Bad date format")


class DateTimeType(SingleInputType):
    widget = "singleline-text"

    def value_from_raw(self, raw):
        if raw.value is None:
            return raw.missing_value("Missing datetime")

        # The previous version of this code allowed a timezone name, followed by a zone
        # offset. In that case the zone name would be ignored (unless the combined zone
        # name, including the offset, matched an IANA zone key). For the sake of
        # backwards compatibility we do the same here.
        m = re.match(
            r"""
            (?P<datetime>
                \d{4} - \d\d? - \d\d?  # YY-MM-DD
                \s+ \d\d? : \d\d (?P<seconds> :\d\d )? # HH:MM[:SS]
            )
            (?: \s+
                (?P<timezone>
                    # Long timezone keys, and — on Windows — names containing
                    # certain characters (those that are not allowed in filenames)
                    # give zoneinfo gas.
                    # https://github.com/python/cpython/issues/96463
                    [^<>:"|?*\x00-\x1f]{,100}?
                    (?P<zoneoffset> [-+] \d\d (?: :? \d\d ){1,2} )?  # ±HHMM[SS]
                )
            )?
            \Z""",
            raw.value.strip(),
            re.DOTALL | re.VERBOSE,
        )
        if m is None:
            return raw.bad_value("Bad datetime format")
        timezone, zoneoffset = m.group("timezone", "zoneoffset")
        tz = None
        if timezone is not None:
            try:
                tz = get_timezone(timezone)
                zoneoffset = None
            # NB: under py39 on windows with tzdata, zoneinfo.ZoneInfo can raise
            # ValueError
            except (LookupError, ValueError):
                if zoneoffset is None:
                    return raw.bad_value(f"Unknown timezone {timezone!r}")

        dt = m["datetime"]
        fmt = "%Y-%m-%d %H:%M"
        if m["seconds"] is not None:
            fmt += ":%S"
        if zoneoffset is not None:
            dt += f" {zoneoffset}"
            fmt += " %z"
        try:
            result = datetime.strptime(dt, fmt)
        except ValueError:
            return raw.bad_value("Invalid datetime")

        if tz is None:
            return result

        # as of babel 2.12, get_timezone can return either a pytz timezone
        # or a zoneinfo timezone
        assert result.tzinfo is None
        if hasattr(tz, "localize"):  # pytz
            return tz.localize(result)
        return result.replace(tzinfo=tz)



import inspect
import json
import os
import re
from pathlib import Path

import pytest

from lektor.builder import Builder
from lektor.cli import cli
from lektor.devserver import run_server
from lektor.project import Project
from lektor.publisher import publish


def test_build_abort_in_existing_nonempty_dir(project_cli_runner):
    os.mkdir("build_dir")
    with open("build_dir/test", "w", encoding="utf-8"):
        pass
    result = project_cli_runner.invoke(cli, ["build", "-O", "build_dir"], input="n\n")
    assert "Aborted!" in result.output
    assert result.exit_code == 1


def test_build_continue_in_existing_nonempty_dir(project_cli_runner):
    os.mkdir("build_dir")
    with open("build_dir/test", "w", encoding="utf-8"):
        pass
    result = project_cli_runner.invoke(cli, ["build", "-O", "build_dir"], input="y\n")
    assert "Finished prune" in result.output
    assert result.exit_code == 0


def test_alias(project_cli_runner):
    result = project_cli_runner.invoke(cli, ["pr"])  # short for 'project-info'
    assert result.exit_code == 0
    assert "Name: Demo Project" in result.output


def test_dev_cmd_alias(isolated_cli_runner):
    result = isolated_cli_runner.invoke(cli, ["dev", "s"])  # short for 'shell'
    assert result.exit_code == 2
    assert "Error: Could not automatically discover project" in result.output


def test_alias_multiple_matches(project_cli_runner):
    result = project_cli_runner.invoke(
        cli, ["p"]
    )  # short for 'project-info' & 'plugins'
    assert result.exit_code == 2
    assert "Error: Too many matches" in result.output


def test_alias_no_matches(project_cli_runner):
    result = project_cli_runner.invoke(cli, ["z"])
    assert result.exit_code == 2
    assert "Error: No such command" in result.output


def test_build_no_project(isolated_cli_runner):
    result = isolated_cli_runner.invoke(cli, ["build"])
    assert result.exit_code == 2
    assert "Could not automatically discover project." in result.output


def test_build(project_cli_runner):
    result = project_cli_runner.invoke(cli, ["build"])
    assert (
        "files or folders already exist" not in result.output
    )  # No warning on fresh build
    assert result.exit_code == 0
    start_matches = re.findall(r"Started build", result.output)
    assert len(start_matches) == 1
    finish_matches = re.findall(r"Finished build in \d+\.\d{2} sec", result.output)
    assert len(finish_matches) == 1

    # rebuild
    result = project_cli_runner.invoke(cli, ["build"])
    assert (
        "files or folders already exist" not in result.output
    )  # No warning on repeat build
    assert result.exit_code == 0


def test_build_extra_flag(project_cli_runner, mocker):
    mock_builder = mocker.patch("lektor.builder.Builder")
    mock_builder.return_value.build_all.return_value = 0
    result = project_cli_runner.invoke(cli, ["build", "-f", "webpack"])
    assert result.exit_code == 0
    assert mock_builder.call_args[1]["extra_flags"] == ("webpack",)


def test_deploy_extra_flag(project_cli_runner, mocker):
    mock_publish = mocker.patch("lektor.publisher.publish")
    result = project_cli_runner.invoke(cli, ["deploy", "-f", "draft"])
    assert result.exit_code == 0
    assert mock_publish.call_args[1]["extra_flags"] == ("draft",)


@pytest.fixture
def project_info_data(project_cli_runner):
    tree_dir = os.getcwd()
    project = Project.from_path(tree_dir)
    return {
        "name": "Demo Project",
        "project_file": os.path.join(tree_dir, "Website.lektorproject"),
        "tree": tree_dir,
        # punt on computing these independently
        "output_path": project.get_output_path(),
        "package_cache": str(project.get_package_cache_path()),
    }


def test_project_info(project_cli_runner, project_info_data):
    result = project_cli_runner.invoke(cli, ["project-info"])
    for heading, key in [
        ("Name", "name"),
        ("File", "project_file"),
        ("Tree", "tree"),
        ("Output", "output_path"),
        ("Package Cache", "package_cache"),
    ]:
        assert f"{heading}: {project_info_data[key]}\n" in result.stdout


@pytest.mark.parametrize(
    "flag",
    ["--name", "--project-file", "--tree", "--output-path", "--package-cache"],
)
def test_project_info_path_flags(project_cli_runner, flag, project_info_data):
    info_key = flag.lstrip("-").replace("-", "_")
    result = project_cli_runner.invoke(cli, ["project-info", flag])
    assert result.exit_code == 0
    assert result.stdout.rstrip() == project_info_data[info_key]


def test_project_info_json(project_cli_runner):
    project = Project.from_path(os.getcwd())
    result = project_cli_runner.invoke(cli, ["project-info", "--json"])
    assert json.loads(result.stdout) == project.to_json()


@pytest.fixture
def deployable_project_data(scratch_project_data):
    project_file = next(scratch_project_data.glob("*.lektorproject"))
    with project_file.open("a") as fp:
        fp.write(
            inspect.cleandoc(
                """[servers.default]
                name = Default
                target = rsync://example.net/
                """
            )
        )
    return scratch_project_data


@pytest.mark.parametrize(
    "subcommand, to_mock, param_name",
    [
        ("build", Builder, "destination_path"),
        ("clean", Builder, "destination_path"),
        ("deploy", publish, "output_path"),
        ("server", run_server, "output_path"),
    ],
)
def test_build_output_path_relative_to_cwd(
    project_cli_runner,
    deployable_project_data,
    mocker,
    subcommand,
    to_mock,
    param_name,
):
    mock = mocker.patch(f"{to_mock.__module__}.{to_mock.__qualname__}", autospec=True)
    args = [
        f"--project={deployable_project_data}",
        subcommand,
        "--output-path=htdocs",
    ]
    project_cli_runner.invoke(cli, args, input="y\n")

    args, kwargs = mock.call_args
    bound_args = inspect.signature(to_mock).bind(*args, **kwargs).arguments
    output_path = bound_args[param_name]
    assert output_path == str(Path.cwd().resolve() / "htdocs")



import inspect
import shutil
from pathlib import Path

import pytest

from lektor.assets import Directory
from lektor.assets import File
from lektor.assets import get_asset
from lektor.assets import get_asset_root
from lektor.project import Project


def write_text(path, text):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(inspect.cleandoc(text))


@pytest.fixture(scope="module")
def project_path(tmp_path_factory, data_path):
    """Make our own private copy of the demo-project"""
    demo = data_path / "demo-project"
    path = tmp_path_factory.mktemp("test_serve") / "demo-project"

    shutil.copytree(demo, path)

    write_text(path / "assets/TEST.TXT", "Text file.\n")
    return path


@pytest.fixture
def pad(project_path, save_sys_path):
    return Project.from_path(project_path).make_env().new_pad()


@pytest.fixture(params=["/", "/static", "/static/demo.css", "/TEST.TXT"])
def asset_path(request):
    return request.param


@pytest.fixture
def asset(pad, asset_path):
    return pad.get_asset(asset_path)


@pytest.mark.parametrize(
    "parent_path, child_name",
    [
        (None, "static"),
        ("/static", "demo.css"),
    ],
)
def test_get_asset(pad, parent_path, child_name):
    parent = pad.get_asset(parent_path) if parent_path is not None else None
    with pytest.deprecated_call(match=r"\bget_asset\b.*\bdeprecated\b") as warnings:
        assert get_asset(pad, child_name, parent=parent).name == child_name
    assert all(warning.filename == __file__ for warning in warnings)


def test_asset_source_filename(asset, pad, asset_path):
    expected = Path(pad.env.root_path, "assets", asset_path.lstrip("/"))
    assert asset.source_filename == str(expected)


@pytest.mark.parametrize(
    "asset_path, url_path",
    [
        ("/", "/"),
        ("/static", "/static/"),
        ("/static/demo.css", "/static/demo.css"),
        ("/TEST.TXT", "/TEST.txt"),
    ],
)
def test_asset_url_path(asset, url_path):
    assert asset.url_path == url_path


@pytest.mark.parametrize(
    "asset_path, expected",
    [
        ("/", "/"),
        ("/static", "/static/"),
        ("/static/demo.css", None),
        ("/TEST.TXT", None),
    ],
)
def test_asset_url_content_path(asset, expected):
    assert asset.url_content_path == expected


@pytest.mark.parametrize(
    "asset_path, artifact_name",
    [
        ("/", "/"),
        ("/static", "/static"),
        ("/static/demo.css", "/static/demo.css"),
        ("/TEST.TXT", "/TEST.txt"),
    ],
)
def test_asset_artifact_name(asset, artifact_name):
    assert asset.artifact_name == artifact_name


@pytest.mark.parametrize(
    "asset_path, child_names",
    [
        ("/dir_with_index_html", {"index.html"}),
        ("/static", {"demo.css"}),
        ("/static/demo.css", set()),
    ],
)
def test_asset_children(asset, child_names):
    assert {child.name for child in asset.children} == child_names


@pytest.mark.parametrize("asset_path", ["/static"])
def test_asset_children_no_children_if_dir_unreadable(asset):
    asset._paths = tuple(
        path.with_name(path.name + "-missing") for path in asset._paths
    )
    assert len(set(asset.children)) == 0


@pytest.mark.parametrize(
    "asset_path, name, child_name",
    [
        ("/", "empty", "empty"),
        ("/", "missing", None),
        ("/", "foo-prefix-makes-me-excluded", None),
        (
            "/",
            "_include_me_despite_underscore",
            "_include_me_despite_underscore",
        ),
        ("/static", "demo.css", "demo.css"),
        ("/static/demo.css", "x", None),
        ("/empty", "demo.css", None),
        # Invalid child names
        ("/static", ".", None),
        ("/", "", None),
    ],
)
def test_asset_get_child(asset, name, child_name):
    if child_name is None:
        assert asset.get_child(name) is None
    else:
        assert asset.get_child(name).name == child_name


def test_asset_get_child_from_url_param_deprecated(asset):
    with pytest.deprecated_call(match=r"\bfrom_url\b.*\bignored\b") as warnings:
        asset.get_child("name", from_url=True)
    assert all(warning.filename == __file__ for warning in warnings)


@pytest.mark.parametrize(
    "asset_path, url_path, expected",
    [
        ("/", ("static",), "/static"),
        ("/", ("static", "demo.css"), "/static/demo.css"),
        ("/", ("static", "demo.css", "parent-not-a-dir"), None),
        ("/", ("missing", "demo.css"), None),
        ("/", ("foo-prefix-makes-me-excluded",), None),
        ("/", ("foo-prefix-makes-me-excluded", "static"), None),
        ("/static", ("demo.css",), "/static/demo.css"),
        ("/", ("TEST.txt",), "/TEST.txt"),
    ],
)
def test_resolve_url_path(asset, url_path, expected):
    if expected is None:
        assert asset.resolve_url_path(url_path) is None
    else:
        assert asset.resolve_url_path(url_path).artifact_name == expected


@pytest.mark.parametrize(
    "asset_path, expected",
    [
        ("/", "<Directory '/'>"),
        ("/static", "<Directory '/static'>"),
        ("/static/demo.css", "<File '/static/demo.css'>"),
        ("/TEST.TXT", "<File '/TEST.txt'>"),
    ],
)
def test_asset_repr(asset, expected):
    assert repr(asset) == expected


@pytest.fixture
def asset_paths(tmp_path):
    paths = tmp_path / "assets1", tmp_path / "assets2"
    for path in paths:
        path.mkdir()
    return paths


@pytest.fixture
def asset_root(pad, asset_paths):
    return get_asset_root(pad, asset_paths)


def test_directory_merges_subdirectories(asset_root, asset_paths):
    for n, path in enumerate(asset_paths):
        subdir = path / "subdir"
        subdir.mkdir()
        subdir.joinpath(f"file{n}").touch()

    subdir_asset = asset_root.get_child("subdir")
    child_names = [child.name for child in subdir_asset.children]
    child_names.sort()
    assert child_names == ["file0", "file1"]


def test_directory_file_shadows_directory(asset_root, asset_paths):
    for n, path in enumerate(asset_paths):
        child_path = path / "child"
        if n == 0:
            child_path.touch()
        else:
            child_path.mkdir()

    children = list(asset_root.children)
    assert len(children) == 1
    assert isinstance(children[0], File)
    assert children[0].name == "child"


def test_directory_directory_conflicts_with_file(asset_root, asset_paths):
    for n, path in enumerate(asset_paths):
        child_path = path / "child"
        if n == 0:
            child_path.mkdir()
        else:
            child_path.touch()

    children = list(asset_root.children)
    assert len(children) == 1
    assert all(isinstance(child, Directory) for child in children)



from __future__ import annotations

import os
from collections.abc import Generator
from typing import Any
from typing import TYPE_CHECKING

import watchfiles

from lektor.utils import get_cache_dir


if TYPE_CHECKING:
    from _typeshed import StrPath

    from lektor.environment import Environment


def watch_project(
    env: Environment, output_path: StrPath, **kwargs: Any
) -> Generator[set[watchfiles.FileChange]]:
    """Watch project source files for changes.

    Returns an generator that yields sets of changes as they are noticed.

    Changes to files within ``output_path`` are ignored, along with other files
    deemed not to be Lektor source files.

    """
    watch_paths = [env.root_path, env.project.project_file, *env.theme_paths]
    ignore_paths = [os.path.abspath(p) for p in (get_cache_dir(), output_path)]
    watch_filter = WatchFilter(env, ignore_paths=ignore_paths)

    return watchfiles.watch(*watch_paths, watch_filter=watch_filter, **kwargs)


class WatchFilter(watchfiles.DefaultFilter):
    def __init__(self, env: Environment, **kwargs: Any):
        super().__init__(**kwargs)
        self.env = env

    def __call__(self, change: watchfiles.Change, path: str) -> bool:
        if self.env.is_uninteresting_source_name(os.path.basename(path)):
            return False
        return super().__call__(change, path)



from lektor.constants import PRIMARY_ALT
from lektor.i18n import get_i18n_block
from lektor.types.base import Type


class FakeType(Type):
    def value_from_raw(self, raw):
        return None

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = Type.to_json(self, pad, record, alt)
        rv["is_fake_type"] = True
        return rv


class LineType(FakeType):
    widget = "f-line"


class SpacingType(FakeType):
    widget = "f-spacing"


class InfoType(FakeType):
    widget = "f-info"


class HeadingType(FakeType):
    widget = "f-heading"

    def to_json(self, pad, record=None, alt=PRIMARY_ALT):
        rv = FakeType.to_json(self, pad, record, alt)
        rv["heading_i18n"] = get_i18n_block(self.options, "heading")
        return rv



from __future__ import annotations

import inspect
import os
import shutil
import sys
from pathlib import Path
from urllib.parse import parse_qsl
from urllib.parse import urljoin
from urllib.parse import urlparse

import flask
import pytest
from werkzeug.exceptions import NotFound

from lektor.admin.context import LektorContext
from lektor.admin.modules import livereload
from lektor.admin.modules import serve
from lektor.admin.webui import LektorApp
from lektor.admin.webui import LektorInfo
from lektor.assets import Asset
from lektor.builder import Artifact
from lektor.buildfailures import FailureController
from lektor.db import Record
from lektor.environment import Environment
from lektor.project import Project


@pytest.fixture(scope="session")
def dummy_app():
    app = flask.Flask("lektor.admin")
    app.config["SERVER_NAME"] = "example.org"
    return app


@pytest.fixture
def dummy_app_context(dummy_app):
    with dummy_app.app_context():
        yield


def make_tooldrawer_config(
    edit_url: str | None = None, artifact: Artifact | None = None
) -> serve.TooldrawerConfig:
    return serve.TooldrawerConfig(
        editUrl=edit_url,
        livereloadConfig=serve.LivereloadConfig.from_artifact(artifact),
    )


@pytest.mark.parametrize(
    "html_text, expect_at_tail",
    [
        ("<html><head></head><body></body></html>", False),
        ("<html>\n<head>\n  </head>\n<body>\n</body>\n</html>", False),
        ("<html><head></ head  ><body></body></html>", False),
        ("<html><head></HeAd><body></body></html>", False),
        ("<html></html>", True),
        ("<html><header></header><body></body></html>", True),
    ],
)
@pytest.mark.usefixtures("dummy_app_context")
def test_inject_tooldrawer(html_text, expect_at_tail):
    eof_mark = "---EOF---"
    config = make_tooldrawer_config("http://example.com/EDIT_URL")
    rewritten = serve._inject_tooldrawer(f"{html_text}{eof_mark}".encode(), config)
    html, _, tail = rewritten.decode("utf-8").rpartition(eof_mark)
    if expect_at_tail:
        assert config.editUrl in tail
        assert config.editUrl not in html
    else:
        assert config.editUrl not in tail
        assert config.editUrl in html


@pytest.mark.usefixtures("dummy_app_context")
def test_inject_tooldrawer_adds_livereload(dummy_app, make_dummy_artifact):
    # Exercises a bug: the injected HTML includes backslash escapes within JSON
    # strings. re.sub(pat, repl, s) doesn't work in that case, as it treats backslashes
    # in repl specially.  This results in an exception: "re.error: bad escape \u at
    # position [...]"
    dummy_app.register_blueprint(livereload.bp)
    artifact = make_dummy_artifact(artifact_name="ARTIFACT_NAME")
    config = make_tooldrawer_config("EDIT_URL", artifact)
    assert b"ARTIFACT_NAME" in serve._inject_tooldrawer(
        html=b"", tooldrawer_config=config
    )


@pytest.mark.usefixtures("dummy_app_context")
def test_inject_tooldrawer_unicode_escapes():
    config = make_tooldrawer_config("http://example.com/EDIT_URL?path=/foo&alt=de")
    assert b"TOOLDRAWER_CONFIG" in serve._inject_tooldrawer(b"", config)


@pytest.fixture
def make_dummy_artifact(tmp_path):
    def make_dummy_artifact(
        artifact_name="test",
        content="<html><head></head><body></body></html>",
    ):
        path = tmp_path / f"{artifact_name}.html"
        if content is not None:
            path.write_text(content)
        return Artifact(
            build_state=None,  # bogus
            artifact_name=artifact_name,
            dst_filename=str(path),
            sources=[],
        )

    return make_dummy_artifact


@pytest.mark.usefixtures("dummy_app_context")
def test_send_html_for_editing(tmp_path, make_dummy_artifact):
    artifact = make_dummy_artifact()
    config = make_tooldrawer_config("EDIT_URL", artifact)
    resp = serve._send_html_for_editing(artifact, config)
    assert resp.mimetype == "text/html"
    # pylint: disable=unsupported-membership-test
    assert "EDIT_URL" in resp.get_data(True)


@pytest.mark.usefixtures("dummy_app_context")
def test_send_html_for_editing_etag_depends_on_edit_url(tmp_path, make_dummy_artifact):
    artifact = make_dummy_artifact()
    resp1 = serve._send_html_for_editing(artifact, make_tooldrawer_config("EDIT_URL1"))
    resp2 = serve._send_html_for_editing(artifact, make_tooldrawer_config("EDIT_URL2"))
    assert resp2.headers["ETag"] != resp1.headers["ETag"]
    resp3 = serve._send_html_for_editing(artifact, make_tooldrawer_config("EDIT_URL1"))
    assert resp3.headers["ETag"] == resp1.headers["ETag"]


@pytest.mark.usefixtures("dummy_app_context")
def test_send_html_for_editing_etag_depends_on_artifact_name(
    tmp_path, make_dummy_artifact
):
    artifact1 = make_dummy_artifact("test1")
    artifact2 = make_dummy_artifact("test2")
    resp1 = serve._send_html_for_editing(
        artifact1, make_tooldrawer_config("EDIT_URL", artifact1)
    )
    resp2 = serve._send_html_for_editing(
        artifact1, make_tooldrawer_config("EDIT_URL", artifact2)
    )
    assert resp2.headers["ETag"] != resp1.headers["ETag"]
    resp3 = serve._send_html_for_editing(
        artifact1, make_tooldrawer_config("EDIT_URL", artifact1)
    )
    assert resp3.headers["ETag"] == resp1.headers["ETag"]


def test_send_html_for_editing_raises_404(tmp_path, make_dummy_artifact):
    artifact = make_dummy_artifact(content=None)
    with pytest.raises(NotFound):
        serve._send_html_for_editing(artifact, "EDIT_URL")


@pytest.mark.parametrize(
    "filename, mimetype",
    [
        ("junk.html", "text/html"),
        ("test.HTM", "text/html"),
        ("test.txt", "text/plain"),
        ("test.foo", "application/octet-stream"),
    ],
)
def test_deduce_mimetype(filename, mimetype):
    assert serve._deduce_mimetype(filename) == mimetype


def test_checked_send_file(tmp_path, dummy_app):
    filename = tmp_path / "file"
    filename.write_text("content")

    with dummy_app.test_request_context():
        resp = serve._checked_send_file(filename, "image/png")
    assert resp.headers["Content-Type"] == "image/png"


def test_checked_send_file_raises_404(tmp_path, dummy_app):
    with dummy_app.test_request_context():
        with pytest.raises(NotFound):
            serve._checked_send_file(tmp_path / "missing.txt", "text/plain")


################################################################


def write_text(path, text):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(inspect.cleandoc(text))


@pytest.fixture(scope="module")
def project_path(tmp_path_factory, data_path):
    """Make our own private copy of the demo-project"""
    demo = data_path / "demo-project"
    path = tmp_path_factory.mktemp("test_serve") / "demo-project"

    shutil.copytree(demo, path)

    # Add a page which will fail to build
    write_text(
        path / "content/extra/build-failure/contents.lr",
        """
        _model: page
        ---
        title: Build Failure
        ---
        _template: missing.html
        """,
    )
    # Add a page which will build to a *.txt file
    write_text(
        path / "content/extra/test.txt/contents.lr",
        """
        _model: page
        ---
        title: Text File
        """,
    )
    return path


@pytest.fixture
def output_path(tmp_path):
    output_path = tmp_path / "output_path"
    output_path.mkdir()
    return output_path


@pytest.fixture(params=[None, ["FLAG"]])
def flag(request):
    return request.param


@pytest.fixture
def app(output_path, project_path, flag):
    project = Project.from_path(project_path)
    env = Environment(project, load_plugins=False)
    lektor_info = LektorInfo(env, output_path, extra_flags=flag)
    app = LektorApp(lektor_info)
    app.register_blueprint(serve.bp, url_prefix="/")
    app.add_url_rule("/ADMIN/EDIT", "url.edit", build_only=True)
    return app


################################################################


class TestArtifactServer:
    # pylint: disable=no-self-use

    @pytest.fixture
    def a_s(self, app):
        lektor_context = LektorContext._make(app.lektor_info)
        return serve.ArtifactServer(lektor_context)

    @pytest.fixture
    def pad(self, app):
        return app.lektor_info.env.new_pad()

    @pytest.fixture
    def failure_controller(self, pad, output_path):
        return FailureController(pad, output_path)

    @pytest.fixture
    def failed_artifact(self, mocker, failure_controller):
        artifact_name = "failed/index.html"
        try:
            raise RuntimeError("Failure")
        except Exception:
            failure_controller.store_failure(artifact_name, sys.exc_info())
        try:
            yield mocker.NonCallableMock(name="Artifact", artifact_name=artifact_name)
        finally:
            failure_controller.clear_failure(artifact_name)

    @pytest.mark.parametrize(
        "url_path, source_path",
        [
            ("", "/"),
            ("index.html", "/"),
            ("de/", "/"),
            ("de/index.html", "/"),
            ("extra/", "/extra"),
            ("extra/index.html", "/extra"),
            ("de/extra/", "/extra"),
            ("de/extra/index.html", "/extra"),
            ("extra/long/path/", "/extra/slash-slug"),
            ("blog", "/blog@1"),
            ("blog/2015/12/post1/", "/blog/post1"),
            ("de/blog/", "/blog@1"),
            ("extra/container", "/extra/container"),
        ],
    )
    def test_resolve_url_path(self, a_s, url_path, source_path):
        source = a_s.resolve_url_path(url_path)
        assert isinstance(source, Record)
        assert source.path == source_path

    @pytest.mark.parametrize(
        "url_path",
        [
            "static/demo.css",
            "dir_with_index_html/",
            "dir_with_index_htm/",
            "static/",
        ],
    )
    def test_resolve_url_path_to_asset(self, a_s, url_path, project_path):
        asset_path = project_path / "assets"
        source = a_s.resolve_url_path(url_path)
        assert isinstance(source, Asset)
        assert Path(source.source_filename) == asset_path / url_path

    @pytest.mark.parametrize(
        "url_path",
        [
            "missing",
            "dir_with_index_html/index.htm",
            "dir_with_index_htm/index.html",
            "static/index.html",
        ],
    )
    def test_resolve_url_path_raises_404(self, a_s, url_path):
        with pytest.raises(NotFound):
            a_s.resolve_url_path(url_path)

    @pytest.mark.parametrize(
        "dir_name, index_name",
        [
            ("dir_with_index_html", "index.html"),
            ("dir_with_index_htm", "index.htm"),
        ],
    )
    def test_resolve_directory_index(self, a_s, pad, dir_name, index_name):
        directory = pad.asset_root.get_child(dir_name)
        index = a_s.resolve_directory_index(directory)
        assert index.name == index_name

    def test_resolve_directory_index_raises_404(self, a_s, pad):
        directory = pad.asset_root.get_child("static")
        with pytest.raises(NotFound):
            a_s.resolve_directory_index(directory)

    @pytest.mark.parametrize(
        "url_path, artifact_name, failing",
        [
            ("de/extra/", "de/extra/index.html", False),
            ("dir_with_index_html/", None, False),
            ("dir_with_index_htm/", None, False),
            ("dir_with_index_html/index.html", "dir_with_index_html/index.html", False),
            ("dir_with_index_htm/index.htm", "dir_with_index_htm/index.htm", False),
            ("static/", None, False),
            ("/extra/build-failure", "extra/build-failure/index.html", True),
            ("/extra/file.ext", "extra/file.ext", False),
        ],
    )
    def test_build_primary_artifact(self, a_s, url_path, artifact_name, failing):
        source = a_s.resolve_url_path(url_path)
        assert source is not None
        if artifact_name is None:
            with pytest.raises(NotFound):
                a_s.build_primary_artifact(source)
        else:
            artifact, failure = a_s.build_primary_artifact(source)
            assert artifact.artifact_name == artifact_name
            assert (failure is not None) == failing

    def test_build_primary_artifact_raises_404(self, a_s, pad):
        source = pad.get("/extra/container")  # has _hidden = yes
        assert source is not None
        with pytest.raises(NotFound):
            a_s.build_primary_artifact(source)

    @pytest.mark.parametrize("edit_url", [None, "EDIT_URL"])
    def test_handle_build_failure(self, app, a_s, mocker, edit_url):
        failure = mocker.NonCallableMock(
            name="BuildFailure",
            data={k: k.upper() for k in ("artifact", "exception", "traceback")},
        )
        with app.test_request_context("failing/"):
            resp = a_s.handle_build_failure(failure, make_tooldrawer_config(edit_url))
        assert resp.status_code == 200
        assert "TRACEBACK" in resp.get_data(True)
        if edit_url is not None:
            assert edit_url in resp.get_data(True)

    @pytest.mark.parametrize(
        "path, kw, expect",
        [
            ("/blog", {"page_num": 1}, {("path", "/blog")}),
            ("/extra", {"alt": "de"}, {("path", "/extra"), ("alt", "de")}),
            ("/extra", {"alt": "en"}, {("path", "/extra")}),
        ],
    )
    def test_get_edit_url(self, a_s, app, pad, path, kw, expect):
        source = pad.get(path, **kw)
        with app.test_request_context(path):
            edit_url = a_s.get_edit_url(source)
        if expect is None:
            assert edit_url is None
        else:
            url = urlparse(edit_url)
            assert url.path == "/ADMIN/EDIT"
            assert set(parse_qsl(url.query)) == expect

    @pytest.mark.parametrize(
        "url_path, location",
        [
            ("extra", "extra/"),
            ("dir_with_index_html", "dir_with_index_html/"),
        ],
    )
    def test_serve_artifact_adds_slash(self, a_s, app, url_path, location):
        with app.test_request_context(f"/{url_path}"):
            resp = a_s.serve_artifact(url_path)
        assert resp.status_code in (301, 308)
        assert resp.location == location

    @pytest.mark.parametrize(
        "url_path, mimetype, is_editable",
        [
            ("projects/coffee/", "text/html", True),  # Page
            ("extra/test.txt", "text/plain", False),  # non-HTML Page
            ("hello.txt", "text/plain", False),  # Attachment
            # Check that build-failure report has edit pencil
            ("extra/build-failure/", "text/html", True),  # Failing build
            # Asset file
            ("static/demo.css", "text/css", False),
            # Page with hidden parent
            ("extra/container/a/", "text/html", True),
            # Asset file with hidden parent
            ("extra/container/hello.txt", "text/plain", False),
            # Asset directories with index.{htm,html}
            ("dir_with_index_html/", "text/html", False),
            ("dir_with_index_htm/", "text/html", False),
            ("dir_with_index_html/index.html", "text/html", False),
            ("dir_with_index_htm/index.htm", "text/html", False),
        ],
    )
    def test_serve_artifact_serves_artifact(
        self, a_s, app, url_path, mimetype, is_editable
    ):
        with app.test_request_context(f"/{url_path}"):
            resp = a_s.serve_artifact(url_path)
            assert resp.status_code == 200
            assert resp.mimetype == mimetype
            data = b"".join(resp.get_app_iter(flask.request.environ)).decode("utf-8")
        assert ("/ADMIN/EDIT?" in data) == is_editable

    @pytest.mark.parametrize(
        "url_path",
        [
            "test@192.jpg",  # sub-artifact — no resolvable to source object
            "extra/container/",  # hidden page
            "static/",  # Asset directory without index.html
            "dir_with_index_html/index.htm",
            "dir_with_index_htm/index.html",
        ],
    )
    def test_serve_artifact_raises_404(self, a_s, app, url_path):
        with app.test_request_context(f"/{url_path}"):
            with pytest.raises(NotFound):
                a_s.serve_artifact(url_path)


################################################################


def test_serve_artifact(app):
    with app.test_request_context("/static/demo.css"):
        resp = serve.serve_artifact("static/demo.css")
        assert resp.status_code == 200
        assert resp.mimetype == "text/css"


def test_serve_file(output_path, app):
    filename = output_path / "test.txt"
    filename.write_text("content")
    with app.test_request_context("/test.txt"):
        resp = serve.serve_file("test.txt")
        content = b"".join(resp.get_app_iter(flask.request.environ))
    assert resp.mimetype == "text/plain"
    assert content == b"content"


@pytest.mark.parametrize("output_path", [Path("relative")])
def test_serve_file_with_relative_output_path(output_path, app, tmp_path):
    # This excercises a bug having to do with serving files when
    # Lektor is given a relative output directory.
    #
    # E.g. via `lektor server -O outdir`
    #
    save_cwd = os.getcwd()
    os.chdir(tmp_path)
    try:
        output_path.mkdir()
        test_serve_file(output_path, app)
    finally:
        os.chdir(save_cwd)


@pytest.mark.parametrize("index_html", ["index.html", "index.htm"])
def test_serve_file_dir_handling(output_path, app, index_html):
    filename = output_path / "adir" / index_html
    filename.parent.mkdir()
    filename.write_text("index")

    with app.test_request_context("/adir"):
        resp = serve.serve_file("adir")
    assert resp.status_code in (301, 308)
    assert resp.location == "adir/"

    with app.test_request_context("/adir/"):
        resp = serve.serve_file("adir/")
        content = b"".join(resp.get_app_iter(flask.request.environ))
    assert resp.status_code == 200
    assert content == b"index"


@pytest.mark.parametrize(
    "path",
    [
        "missing",
        "example/container",  # hidden page
        "adir/",  # no adir/index.{html,htm} exists
        "adir/../top.txt",  # ".." not allowed in path
        "../adir/index.txt",  # points outside of output_path
        "adir/index.txt/../index.txt",
        "adir/index.txt/../../top.txt",
        "adir/index.txt/../../adir/index.txt",
    ],
)
def test_serve_file_raises_404(output_path, app, path):
    not_an_index = output_path / "adir/index.txt"
    not_an_index.parent.mkdir()
    not_an_index.write_text("non-index")
    output_path.joinpath("top.txt").write_text("top")

    with app.test_request_context(path):
        with pytest.raises(NotFound):
            serve.serve_file(path)


################################################################


@pytest.mark.parametrize(
    "path, status, mimetype, content",
    [
        ("/hello.txt", 200, "text/plain", "Hello I am an Attachment"),
        ("/missing/", 404, "text/html", "The requested URL was not found"),
        ("/extra/container/", 404, "text/html", "Record is hidden"),
    ],
)
def test_serve(app, path, status, mimetype, content):
    with app.test_client() as c:
        resp = c.get(path)
        assert resp.status_code == status
        assert resp.mimetype == mimetype
        assert content in resp.get_data(True)


def test_serve_from_file(app, output_path):
    filename = output_path / "sub-artifact.txt"
    filename.write_text("sub-artifact")
    with app.test_client() as c:
        resp = c.get("/sub-artifact.txt")
        assert resp.status_code == 200
        assert resp.mimetype == "text/plain"


@pytest.mark.parametrize(
    "path_info, base_url, location",
    [
        ("/extra", "http://example.org/pfx/", "http://example.org/pfx/extra/"),
        (
            "/dir_with_index_html?qs",
            "http://localhost/",
            "http://localhost/dir_with_index_html/?qs",
        ),
        (
            "/projects/coffee",
            "http://localhost/pfx/",
            "http://localhost/pfx/projects/coffee/",
        ),
        ("/adir", "http://localhost/", "http://localhost/adir/"),
        ("/adir/bdir", "http://localhost/", "http://localhost/adir/bdir/"),
    ],
)
def test_serve_add_slash_redirect_integration(
    app, output_path, path_info, base_url, location
):
    output_path.joinpath("adir/bdir").mkdir(parents=True)
    with app.test_client() as c:
        resp = c.get(path_info, base_url=base_url)
        assert resp.status_code in (301, 308)
        # Be careful to accept either absolute or relative URLs in resp.location
        # See https://httpwg.org/specs/rfc7231.html#header.location
        request_url = urljoin(base_url, path_info.lstrip("/"))
        assert urljoin(request_url, resp.location) == location


@pytest.fixture
def scratch_app(scratch_env, output_path):
    lektor_info = LektorInfo(scratch_env, output_path)
    app = LektorApp(lektor_info)
    app.register_blueprint(serve.bp, url_prefix="/")
    app.add_url_rule("/ADMIN/EDIT", "url.edit", build_only=True)
    return app


def test_serve_custom_404(scratch_app, scratch_project_data):
    custom_404 = scratch_project_data / "assets/404.html"
    custom_404.parent.mkdir(exist_ok=True)
    custom_404.write_text("custom 404")

    with scratch_app.test_client() as c:
        resp = c.get("/missing.txt")
        assert resp.status_code == 404
        assert resp.mimetype == "text/html"
        assert resp.get_data(True) == "custom 404"



import base64
import codecs
import hashlib
import mimetypes
import os

from lektor.utils import deprecated


class FileContents:
    @deprecated(name="FileContents", version="3.4.0")
    def __init__(self, filename):
        self.filename = filename
        self._md5 = None
        self._sha1 = None
        self._integrity = None
        self._mimetype = mimetypes.guess_type(filename)[0] or "application/octet-stream"

    @property
    def sha1(self):
        self._ensure_hashes()
        return self._sha1

    @property
    def md5(self):
        self._ensure_hashes()
        return self._md5

    @property
    def integrity(self):
        self._ensure_hashes()
        return self._integrity

    @property
    def mimetype(self):
        return self._mimetype

    @property
    def bytes(self):
        try:
            return os.stat(self.filename).st_size
        except OSError:
            return 0

    def as_data_url(self, mediatype=None):
        if mediatype is None:
            mediatype = self.mimetype
        return f"data:{mediatype};base64,{self.as_base64()}"

    def as_text(self):
        with self.open() as f:
            return f.read()

    def as_bytes(self):
        with self.open("rb") as f:
            return f.read()

    def as_base64(self):
        return base64.b64encode(self.as_bytes())

    def open(self, mode="r", encoding=None):
        if mode == "rb":
            return open(self.filename, "rb")
        if mode != "r":
            raise TypeError("Can only open files for reading")
        # pylint: disable=deprecated-method
        return codecs.open(self.filename, encoding=encoding or "utf-8")

    def _ensure_hashes(self):
        if self._md5 is not None:
            return
        with self.open("rb") as f:
            md5 = hashlib.md5()
            sha1 = hashlib.sha1()
            sha384 = hashlib.sha384()
            while 1:
                chunk = f.read(16384)
                if not chunk:
                    break
                md5.update(chunk)
                sha1.update(chunk)
                sha384.update(chunk)
            self._md5 = md5.hexdigest()
            self._sha1 = sha1.hexdigest()
            self._integrity = "sha384-" + base64.b64encode(sha384.digest()).decode(
                "ascii"
            )

    def __repr__(self):
        return f"<FileContents {self.filename!r} md5={self.md5!r}>"



from __future__ import annotations

import dataclasses
import html.parser
import io
import os
import shutil
from contextlib import contextmanager
from pathlib import Path

import PIL
import pytest
from pytest import approx

from lektor.context import Context
from lektor.db import Image
from lektor.imagetools.thumbnail import _compute_cropbox
from lektor.imagetools.thumbnail import _convert_icc_profile_to_srgb
from lektor.imagetools.thumbnail import _create_artifact
from lektor.imagetools.thumbnail import _create_thumbnail
from lektor.imagetools.thumbnail import _get_thumbnail_url_path
from lektor.imagetools.thumbnail import compute_dimensions
from lektor.imagetools.thumbnail import CropBox
from lektor.imagetools.thumbnail import get_image_info
from lektor.imagetools.thumbnail import ImageSize
from lektor.imagetools.thumbnail import make_image_thumbnail
from lektor.imagetools.thumbnail import Thumbnail
from lektor.imagetools.thumbnail import ThumbnailMode
from lektor.imagetools.thumbnail import ThumbnailParams


HERE = Path(__file__).parent
DEMO_PROJECT = HERE / "../demo-project/content"
ICC_PROFILE_TEST_JPG = DEMO_PROJECT / "icc-profile-test/rgb-to-gbr-test.jpg"
CMYK_ICC_PROFILE_TEST = DEMO_PROJECT / "icc-profile-test/CGATS001Compat-v2-micro.icc"
NONIMAGE_FILE_PATH = Path(__file__)  # we are not an image


def test_ThumbnailParams_unrecognized_format():
    with pytest.raises(ValueError, match="unrecognized format"):
        ThumbnailParams(ImageSize(80, 60), "UNKNOWNFORMAT")


@pytest.mark.parametrize(
    "format, quality, expected",
    [
        ("GIF", None, {"format": "GIF"}),
        ("gif", 99, {"format": "GIF"}),
        ("JPEG", None, {"format": "JPEG", "quality": 85}),
        ("JPEG", 95, {"format": "JPEG", "quality": 95}),
        ("PNG", None, {"format": "PNG", "compress_level": 7}),
        ("PNG", 92, {"format": "PNG", "compress_level": 9}),
        ("PNG", 180, {"format": "PNG", "compress_level": 9}),
        ("PNG", 75, {"format": "PNG", "compress_level": 7}),
        ("PNG", -10, {"format": "PNG", "compress_level": 0}),
    ],
)
def test_ThumbnailParams_get_save_params(format, quality, expected):
    thumbnail_params = ThumbnailParams(ImageSize(10, 10), format, quality)
    assert thumbnail_params.get_save_params() == expected


@pytest.mark.parametrize(
    "format, proposed_ext, expected",
    [
        ("GIF", ".gif", ".gif"),
        ("GIF", ".gIf", ".gIf"),
        ("GIF", ".bad", ".gif"),
        ("GIF", "GIF", ".gif"),
        ("JPEG", ".jpg", ".jpg"),
        ("JPEG", ".JPEG", ".JPEG"),
        ("JPEG", ".j", ".jpeg"),
        ("PNG", ".x", ".png"),
        ("PNG", ".PNG", ".PNG"),
    ],
)
def test_ThumbnailParams_get_ext(format, proposed_ext, expected):
    thumbnail_params = ThumbnailParams(ImageSize(10, 20), format)
    assert thumbnail_params.get_ext(proposed_ext) == expected


@pytest.mark.parametrize(
    "format, width, height, quality, crop, expected",
    [
        ("GIF", 10, 20, None, False, "10x20"),
        ("gif", 10, 20, 92, False, "10x20"),
        ("GIF", 10, 20, None, True, "10x20_crop"),
        ("JPEG", 200, 300, 92, False, "200x300_q92"),
        ("JPEG", 200, 300, None, True, "200x300_crop"),
        ("PNG", 1, 2, 92, True, "1x2_crop_q9"),
        ("PNG", 4, 5, None, False, "4x5"),
    ],
)
def test_ThumbnailParams_get_tag(format, width, height, quality, crop, expected):
    thumbnail_params = ThumbnailParams(ImageSize(width, height), format, quality, crop)
    assert thumbnail_params.get_tag() == expected


@pytest.mark.parametrize(
    "width, height, source_width, source_height, expected",
    [
        (1, None, 100, 200, ImageSize(1, 2)),
        (1, None, 100, 249, ImageSize(1, 2)),
        (1, None, 100, 250, ImageSize(1, 3)),
        (None, 2, 100, 200, ImageSize(1, 2)),
        (10, 10, 100, 200, ImageSize(5, 10)),
        (10, 21, 100, 200, ImageSize(10, 20)),
        # tests from test_images.py::test_dimensions
        # landscape
        (50, 50, 100, 50, ImageSize(50, 25)),
        (100, 20, 100, 50, ImageSize(40, 20)),
        (200, 200, 100, 50, ImageSize(200, 100)),
        (500, 200, 100, 50, ImageSize(400, 200)),
        # test missing dimension
        (50, None, 100, 50, ImageSize(50, 25)),
        (None, 20, 100, 50, ImageSize(40, 20)),
        (200, None, 100, 50, ImageSize(200, 100)),
        (None, 200, 100, 50, ImageSize(400, 200)),
        # test that rounding is half-up
        (49, None, 100, 50, ImageSize(49, 25)),
        (51, None, 100, 50, ImageSize(51, 26)),
        #
        # portrait
        (50, 50, 50, 100, ImageSize(25, 50)),
        (20, 100, 50, 100, ImageSize(20, 40)),
        (200, 200, 50, 100, ImageSize(100, 200)),
        (200, 500, 50, 100, ImageSize(200, 400)),
        #
        (None, 50, 50, 100, ImageSize(25, 50)),
        (20, None, 50, 100, ImageSize(20, 40)),
        (None, 200, 50, 100, ImageSize(100, 200)),
        (200, None, 50, 100, ImageSize(200, 400)),
        #
        (None, 49, 50, 100, ImageSize(25, 49)),
        (None, 51, 50, 100, ImageSize(26, 51)),
    ],
)
def test_compute_dimensions(width, height, source_width, source_height, expected):
    size = compute_dimensions(width, height, source_width, source_height)
    assert size == expected


def test_compute_dimensions_raises_value_error():
    with pytest.raises(ValueError, match="may not both be None"):
        compute_dimensions(None, None, 10, 20)


@pytest.mark.parametrize(
    "size, source_width, source_height, expected",
    [
        (ImageSize(1, 1), 10, 20, CropBox(0, 5, 10, 15)),
        (ImageSize(1, 1), 10, 11, CropBox(0, 0, 10, 10)),
        (ImageSize(1, 1), 20, 10, CropBox(5, 0, 15, 10)),
    ],
)
def test_compute_cropbox(size, source_width, source_height, expected):
    assert _compute_cropbox(size, source_width, source_height) == expected


@pytest.fixture(scope="session")
def dummy_image():
    return PIL.Image.new("RGB", (100, 100), "#999")


def test_convert_icc_profile_to_srgb():
    im = PIL.Image.open(ICC_PROFILE_TEST_JPG)
    # Top center of image is blue before color transform
    assert im.getpixel((im.width // 2, 0)) == approx((0, 0, 255), abs=10)

    _convert_icc_profile_to_srgb(im)
    # Top center of image is red after color transform
    assert im.getpixel((im.width // 2, 0)) == approx((255, 0, 0), abs=10)
    assert "icc_profile" not in im.info


def test_convert_icc_profile_to_srgb_no_profile():
    im = PIL.Image.new("RGB", (100, 100), "#999")
    _convert_icc_profile_to_srgb(im)
    assert "icc_profile" not in im.info


def test_create_thumbnail(dummy_image):
    thumbnail_params = ThumbnailParams(ImageSize(80, 60), "PNG")
    thumb = _create_thumbnail(dummy_image, thumbnail_params)
    assert thumb.width == 80 and thumb.height == 60
    assert thumb.getpixel((40, 30)) == approx((153, 153, 153), abs=5)


def test_create_thumbnail_converts_hsv_to_rgb():
    hsv_image = PIL.Image.new("HSV", (100, 100), (0, 0, 153))
    thumb = _create_thumbnail(hsv_image, ThumbnailParams(ImageSize(80, 60), "JPEG"))
    assert thumb.mode == "RGB"
    assert thumb.getpixel((40, 30)) == approx((153, 153, 153), abs=2)


def test_create_thumbnail_converts_indexed_to_rgb():
    image = PIL.Image.new("P", (100, 100), 0)
    image.putpalette(b"\x10\x20\x30", "RGB")
    thumb = _create_thumbnail(image, ThumbnailParams(ImageSize(80, 60), "JPEG"))
    assert thumb.mode == "RGB"
    assert thumb.getpixel((40, 30)) == approx((16, 32, 48), abs=2)


def test_create_thumbnail_converts_cmyk_to_rgb(dummy_image):
    cmyk_image = dummy_image.convert("CMYK")
    thumb = _create_thumbnail(cmyk_image, ThumbnailParams(ImageSize(80, 60), "JPEG"))
    assert thumb.mode == "RGB"
    assert thumb.getpixel((40, 30)) == approx((153, 153, 153), abs=5)


def test_create_thumbnail_converts_cmyk_to_rgb_via_icc_profile(dummy_image):
    cmyk_image = dummy_image.convert("CMYK")
    cmyk_image.info["icc_profile"] = CMYK_ICC_PROFILE_TEST.read_bytes()
    thumb = _create_thumbnail(cmyk_image, ThumbnailParams(ImageSize(80, 60), "PNG"))
    assert thumb.mode == "RGB"
    assert thumb.getpixel((40, 30)) == approx((162, 148, 145), abs=5)


class DummyArtifact:
    image = None

    @contextmanager
    def open(self, _mode):
        fp = io.BytesIO()
        yield fp
        fp.seek(0)
        self.image = PIL.Image.open(fp)


def test_create_artifact(dummy_jpg_path):
    thumbnail_params = ThumbnailParams(ImageSize(80, 60), "JPEG")
    artifact = DummyArtifact()
    _create_artifact(dummy_jpg_path, thumbnail_params, artifact)
    thumb = artifact.image
    assert thumb.width == 80 and thumb.height == 60
    assert thumb.getpixel((40, 30)) == approx((153, 153, 153), abs=5)


def test_create_artifact_strips_metadata():
    # this image has comment, dpi, exif, icc_profile, and photoshop data
    source_image = ICC_PROFILE_TEST_JPG
    jfif_info_keys = {"jfif", "jfif_density", "jfif_unit", "jfif_version"}
    thumbnail_params = ThumbnailParams(ImageSize(80, 60), "JPEG")
    artifact = DummyArtifact()
    _create_artifact(source_image, thumbnail_params, artifact)
    metadata_keys = set(artifact.image.info) - jfif_info_keys
    assert not metadata_keys


@pytest.mark.parametrize(
    "source_url_path, size, format, expected",
    [
        (
            "/urlpath/foo.bar.JPG",
            ImageSize(10, 20),
            "JPEG",
            "/urlpath/foo.bar@10x20.JPG",
        ),
        ("/urlpath/foo.bar", ImageSize(10, 20), "JPEG", "/urlpath/foo@10x20.jpeg"),
    ],
)
def test_get_thumbnail_url_path(source_url_path, size, format, expected):
    thumbnail_params = ThumbnailParams(size, format)
    assert _get_thumbnail_url_path(source_url_path, thumbnail_params) == expected


@pytest.fixture(scope="session")
def dummy_jpg_path(tmp_path_factory):
    dummy_jpg = tmp_path_factory.mktemp("images") / "dummy.jpg"
    im = PIL.Image.new("RGB", (400, 300), "#999")
    with dummy_jpg.open("wb") as fp:
        im.save(fp, "JPEG")
    return dummy_jpg


@pytest.fixture
def ctx(builder):
    build_state = builder.new_build_state()
    with Context(build_state.new_artifact("dummy-artifact")) as ctx:
        yield ctx


@pytest.mark.parametrize(
    "source_url_path, kwargs, expected_size, thumbnail_url_path",
    [
        # ThumbnailMode.FIT
        ("/test.jpg", {"width": 80, "height": 80}, (80, 60), "/test@80x60.jpg"),
        ("/test.jpg", {"width": 80}, (80, 60), "/test@80x60.jpg"),
        ("/test.jpg", {"height": 90, "quality": 85}, (120, 90), "/test@120x90_q85.jpg"),
        (
            "/test.jpg",
            {"width": 80, "mode": ThumbnailMode.FIT},
            (80, 60),
            "/test@80x60.jpg",
        ),
        # ThumbnailMode.CROP
        (
            "/test.jpg",
            {"width": 80, "height": 80, "mode": ThumbnailMode.CROP},
            (80, 80),
            "/test@80x80_crop.jpg",
        ),
        (
            "/test.jpg",
            {"width": 800, "height": 800, "mode": ThumbnailMode.CROP},
            (800, 800),
            "/test@800x800_crop.jpg",
        ),
        # ThumbnailMode.STRETCH
        (
            "/test.jpg",
            {"width": 80, "height": 80, "mode": ThumbnailMode.STRETCH},
            (80, 80),
            "/test@80x80.jpg",
        ),
        (
            "/test.jpg",
            {"width": 800, "height": 800, "mode": ThumbnailMode.STRETCH},
            (800, 800),
            "/test@800x800.jpg",
        ),
        # explicit upscale
        ("/test.jpg", {"width": 440, "upscale": True}, (440, 330), "/test@440x330.jpg"),
        # explicit upscale=False
        ("/test.jpg", {"width": 440, "upscale": False}, (400, 300), "/test.jpg"),
        # implicit upscale
        ("/test.jpg", {"width": 440}, (400, 300), "/test.jpg"),
    ],
)
def test_make_image_thumbnail(
    ctx, source_url_path, kwargs, expected_size, thumbnail_url_path, dummy_jpg_path
):
    thumbnail = make_image_thumbnail(ctx, dummy_jpg_path, source_url_path, **kwargs)
    assert (thumbnail.width, thumbnail.height) == expected_size
    assert thumbnail.url_path == thumbnail_url_path
    if "@" in thumbnail_url_path:
        assert len(ctx.sub_artifacts) == 1
    else:
        assert len(ctx.sub_artifacts) == 0  # no implicit upscale


@pytest.mark.parametrize(
    "params, match",
    [
        ({}, "at least one of width or height"),
        ({"width": 8, "mode": ThumbnailMode.CROP}, "requires both"),
    ],
)
def test_make_image_thumbnail_invalid_params(ctx, params, match, dummy_jpg_path):
    with pytest.raises(ValueError, match=match):
        make_image_thumbnail(ctx, dummy_jpg_path, "/test.jpg", **params)
    assert len(ctx.sub_artifacts) == 0


def test_make_image_thumbnail_unknown_image_format(ctx, tmp_path):
    with pytest.raises(RuntimeError, match="unknown"):
        make_image_thumbnail(ctx, NONIMAGE_FILE_PATH, "/test.jpg", width=80)
    assert len(ctx.sub_artifacts) == 0


@pytest.fixture
def dummy_svg_file(tmp_path):
    dummy_svg = tmp_path / "dummy.svg"

    def svg_file(**kwargs):
        attrs = " ".join(f'{key}="{value}"' for key, value in kwargs.items())
        dummy_svg.write_text(f'<svg {attrs} xmlns="http://www.w3.org/2000/svg"></svg>')
        return dummy_svg

    return svg_file


def test_make_image_thumbnail_svg(ctx, dummy_svg_file):
    svg_file = dummy_svg_file(width="400px", height="300px")
    thumbnail = make_image_thumbnail(
        ctx, svg_file, "/urlpath/dummy.svg", width=80, height=60
    )
    assert (thumbnail.width, thumbnail.height) == (80, 60)
    assert thumbnail.url_path == "/urlpath/dummy.svg"
    assert len(ctx.sub_artifacts) == 0


def test_make_image_thumbnail_svg_fit_mode_fails_if_missing_dim(ctx, dummy_svg_file):
    svg_file = dummy_svg_file(width="400px")  # missing height
    with pytest.raises(ValueError, match="Cannot determine aspect ratio"):
        make_image_thumbnail(ctx, svg_file, "/urlpath/dummy.svg", width=80, height=60)


def test_Thumbnail_str():
    thumbnail = Thumbnail("/urlpath/image.jpg", 50, 80)
    assert str(thumbnail) == "image.jpg"


class TestArtifactDependencyTracking:
    @staticmethod
    @pytest.fixture
    def scratch_project_data(scratch_project_data):
        # add two identical thumbnails to the page template
        page_html = scratch_project_data / "templates/page.html"
        with page_html.open("a") as fp:
            fp.write(
                """
                {% set im = this.attachments.get('test.jpg') %}
                <img src="{{ im.thumbnail(20) }}">
                <img src="{{ im.thumbnail(20) }}">
                """
            )
        shutil.copy(ICC_PROFILE_TEST_JPG, scratch_project_data / "content/test.jpg")
        return scratch_project_data

    @staticmethod
    def build_page(builder) -> list[str]:
        _, build_state = builder.build(builder.pad.root)
        assert len(build_state.failed_artifacts) == 0
        return [
            os.path.basename(artifact.dst_filename)
            for artifact in build_state.updated_artifacts
        ]

    def test(self, scratch_builder):
        built = self.build_page(scratch_builder)
        assert built == ["index.html", "test@20x20.jpg"]

        # rebuild doesn't rebuild any more artifacts
        Path(scratch_builder.destination_path, "index.html").unlink()
        built = self.build_page(scratch_builder)
        assert built == ["index.html"]

    @pytest.mark.xfail(reason="This will (or should, at least) be fixed by PR #1148")
    def test_racy_source_changes(self, scratch_builder, monkeypatch):
        # Modify source image immediately after the first call to Image.thumbnail().
        # Note that this occurs *before* the thumbnail artifact is built.
        #
        # We are doing this to ensure there are no race conditions in
        # the dependency tracking. Since the source was changed after the thumbnail
        # parameters were computed, the artifact should be updated at the next build.

        def thumbnail_advice(self, *args, **kwargs):
            monkeypatch.undo()
            try:
                return Image.thumbnail(self, *args, **kwargs)
            finally:
                with open(self.attachment_filename, "ab") as fp:
                    fp.write(b"\0")

        monkeypatch.setattr(Image, "thumbnail", thumbnail_advice)

        built = self.build_page(scratch_builder)
        assert built == ["index.html", "test@20x20.jpg"]

        Path(scratch_builder.destination_path, "index.html").unlink()
        built = self.build_page(scratch_builder)
        assert built == ["index.html", "test@20x20.jpg"]


@dataclasses.dataclass
class DemoThumbnail:
    width: int
    height: int
    max_size: int | None = None


@dataclasses.dataclass
class ImgTag:
    src: str
    width: int | None
    height: int | None
    alt: str


class TestFunctional:
    """Remnants of the tests that were in test_images.py"""

    # pylint: disable=no-self-use

    @pytest.fixture
    def thumbnails(self):
        return {
            # original dimensions = 384 x 512
            "test@192x256.jpg": DemoThumbnail(192, 256),
            "test@300x100_crop.jpg": DemoThumbnail(300, 100),
            "test@300x100.jpg": DemoThumbnail(300, 100),
            "test@192x256_q20.jpg": DemoThumbnail(192, 256, 9200),
        }

    @pytest.fixture
    def built_demo_img_tags(self, built_demo):
        def int_val(value):
            if value is None:
                return None
            return int(value)

        imgs = []

        def handle_starttag(tag, attrs):
            if tag == "img":
                a = dict(attrs)
                imgs.append(
                    ImgTag(
                        src=a.get("src"),
                        alt=a.get("alt"),
                        width=int_val(a.get("width")),
                        height=int_val(a.get("height")),
                    )
                )

        parser = html.parser.HTMLParser()
        parser.handle_starttag = handle_starttag
        parser.feed(Path(built_demo, "index.html").read_text("utf-8"))
        parser.close()
        return imgs

    def test_img_tag_dimensions(self, built_demo_img_tags, thumbnails):
        seen = set()

        for img in built_demo_img_tags:
            seen.add(img.src)
            thumb = thumbnails.get(img.src)
            if thumb is not None:
                assert (img.width, img.height) == (thumb.width, thumb.height)
        assert seen >= set(thumbnails)

    def test_thumbnail_dimensions(self, built_demo, thumbnails):
        for name, thumb in thumbnails.items():
            _format, width, height = get_image_info(built_demo / name)
            assert (width, height) == (thumb.width, thumb.height)

    def test_thumbnails_distinct(self, built_demo, thumbnails):
        images = set()
        for name in thumbnails:
            images.add(Path(built_demo / name).read_bytes())
        assert len(images) == len(thumbnails)

    def test_max_size(self, built_demo, thumbnails):
        for name, thumb in thumbnails.items():
            if thumb.max_size is not None:
                assert Path(built_demo / name).stat().st_size < thumb.max_size

    def test_large_thumbnail_returns_original(self, built_demo_img_tags):
        large_thumb = ImgTag(alt="original", src="test.jpg", width=384, height=512)
        assert large_thumb in built_demo_img_tags
